# pyright: reportOperatorIssue = false
from re import escape
from string import digits
from sys import version_info
from typing import Callable
from inspect import signature, Parameter

from .psuedonyms import psuedonyms
from .EZRegex import EZRegex, EZRegexFunc
from functools import partial
from string import Formatter


from .types import EZRegexFunc, EZRegexType, EZRegexDefinition, EZRegexOther, EZRegexParam
# TODO: add typing to all of these

# NOTE: the docstrings are added manually, which I hate. They SHOULD be below the variable
# and auto-parsed, but that's difficult for a number of reasons, one of which being my
# weird, non-standard structure. Adding manually works for now.

# NOTE: when writing mixins, keep in mind that parameters which are None or bools will be passed as-is,
# ints will be cast to strings, and strings will be escaped based on the dialect's _escape_chars. Any
# other types will be auto-cast to strings, special characters will not be escaped, and a warning thrown
# See EZRegex._sanitize_param for more info

def raise_if_empty(param, func, param_name='pattern'):
    if not len(str(param)):
        try:
            func_names = psuedonyms[func]
        except KeyError:
            func_names = False
        raise ValueError(f'Parameter {param_name} in {func} {"(aka " + ", ".join(func_names) + ")" if func_names else ""} cannot be empty')


def imply_pattern_is_cur(func):
    """ If `pattern` is Ellipsis, it will use `cur` instead, and `cur` will be set to an empty string.
        This is useful for functions that want to allow both inline and operator style chaining
        i.e. digit.amt(2) and amt(2, digit)

        NOTE: `pattern` must be a keyword parameter, and it must be the last parameter able to be
        provided as a positional argument. Don't use *args.

        NOTE: must come *after* add_greedy_possessive, not before. Don't ask me why.
    """
    def rtn(*args, pattern=..., cur=..., **kwargs):
        # First, check how many positional only params there are
        # parameters is an ordered mapping
        for num_params, param in enumerate(signature(func).parameters.values()):
            if param.name == 'pattern': break

        # If pattern is provided as a positional arg, make it into a keyword arg
        if len(args) > num_params:
            pattern = args[-1]
            args = args[:-1]

        if pattern is Ellipsis:
            return func(*args, pattern=cur, cur='', **kwargs)

        raise_if_empty(pattern, func.__name__)
        return func(*args, pattern=pattern, cur=cur, **kwargs)
    return rtn

def _parse_any_of_params(*patterns, chars=None, split=None):
    if split and len(patterns) != 1:
        raise ValueError("Please don't specifiy split and pass multiple patterns to anyof")
    elif split:
        patterns = list(patterns[0])
    elif len(patterns) == 1 and split is None and chars is not False:  # None means auto
        chars = True
        patterns = list(patterns[0])
    elif len(patterns) == 1 and split is None:
        patterns = list(patterns[0])
    elif len(patterns) > 1 and chars is None and all(map(lambda s: len(str(s)) == 1, patterns)):
        chars = True

    return chars, patterns

def BaseMixin(*, allow_greedy=False, allow_possessive=False):
    """ The basics of regex syntax. Almost all dialects have these. You almost certainly
        want to inherit from this class, even if you need to overload a few of it's members.
    """

    def add_greedy_possessive(func):
        def rtn(*args, greedy=True, possessive=False, cur=..., **kwargs):
            if not greedy and not allow_greedy:
                raise ValueError('Greedy qualifiers are not allowed in this dialect')
            if possessive and not allow_possessive:
                raise ValueError('Possessive qualifiers are not allowed in this dialect')
            if (not greedy) and possessive:
                raise ValueError('You can\'t be both non-greedy and possessive at the same time')
            return func(*args, cur=cur, **kwargs)
        return rtn

    class _BaseMixin:
        # replacement = None means it can work with both replacement and non-replacement EZRegex chains
        literal = lambda pattern, cur=...: cur + pattern, {'replacement': None, 'docstring':
        """This is a redundant function. You should always be able to use `... + 'stuff'` just as easily as `... + literal('stuff')`

            Args:
                pattern: the pattern to add
        """}

        # Not technically a variable, but it's accepted by EZRegex.__init__(), so it works
        raw = lambda regex, cur=...: cur + regex, {'_is_raw': True, 'replacement': None, 'docstring':
        """ If you already have some regular regex written and you want to incorperate
            it, this will allow you to include it without sanitizing all the backslashes
            and such, which all the other EZRegexs do automatically

            Args:
                regex: the regex to add directly
        """}

        # Positional
        word_boundary      = r'\b'
        not_word_boundary  = r'\B'

        # Literals
        tab                = r'\t'
        space              = r' '
        space_or_tab       = r'[ \t]'
        new_line           = r'\n'
        carriage_return    = r'\r'
        quote              = r'(?:\'|"|`)', {'docstring':
        "Matches ', \", and `"}
        vertical_tab       = r'\v'
        form_feed          = r'\f'
        comma              = r'\,'
        period             = r'\.'
        underscore         = r'_'

        # Not Literals
        not_whitespace     = r'\S'
        not_digit          = r'\D'
        not_word           = r'\W'

        # Catagories
        white_char         = r'\s'
        whitechunk         = r'\s+', {'docstring':
        "A \"chunk\" of whitespace. Just any amount of whitespace together"}
        digit              = r'\d'
        number             = r'\d+', {'docstring':
        "Matches multiple digits next to each other. Does not match negatives or decimals"}
        word               = r'\w+'
        word_char          = r'\w', {'docstring':
        "Matches just a single \"word character\", defined as any letter, number, or _"}
        anything           = r'.', {'docstring':
        "Matches any single character, except a newline. To also match a newline, use literally_anything"}
        chunk              = r'.+', {'docstring':
        "A \"chunk\": Any clump of characters up until the next newline"}
        uppercase          = r'[A-Z]'
        lowercase          = r'[a-z]'
        letter             = r'[A-Za-z]', {'docstring':
        "Matches just a letter -- not numbers or _ like word_char"}
        hex_digit          = r'[0-9a-fA-F]'
        oct_digit          = r'[0-7]'
        # TODO: is there a more formal definition of this or something?
        # NOTE: this could cause problems, as this is the *python* regex escape function, not the
        # current class's.
        # However, getting access to the current class here is sorta impossible, so I'm ignoring it until
        # it starts to cause problems
        punctuation         = r'[' + escape(']`~!@#$%^&*()-_=+[{}\\|;:\'",<.>/?¢]') + r']'
        controller          = r'[\x00-\x1F\x7F]', {'docstring':
        "Matches a metadata ASCII characters"}
        printable           = r'[\x21-\x7E]', {'docstring':
        "Matches printable ASCII characters"}
        printable_and_space = r'[\x20-\x7E]'
        letter_num          = r'[A-Za-z0-9_]'
        unicode             = lambda name, cur=...: fr'\N{name}', {'docstring':
        """ Matches a unicode character by name

            Args:
                name: the name of the unicode character
        """}

        # Premade
        # TODO: a chunk of literally anything/chunk of literally anything except ...
        literally_anything = r'(?:.|\n)', {'docstring':
        "*Any* character, include newline"}
        signed             = r'(?:(?:\-|\+))?\d+', {'docstring':
        "a signed number, including 123, -123, and +123"}
        unsigned           = r'\d+', {'docstring':
        "Same as number. Will not match +123"}
        plain_float        = r'(?:(?:\-|\+))?\d+\.(?:\d+)?', {'docstring':
        "Will match 123.45 and 123."}
        full_float         = r'(?:(?:\-|\+))?\d+\.(?:\d+)?(?:e(?:(?:\-|\+))?\d+)?', {'docstring':
        "Will match plain_float as well as things like 1.23e-10 and 1.23e+10"}
        int_or_float       = r'(?:(?:\-|\+))?\d+\.(?:\d+)?(?:e(?:(?:\-|\+))?\d+)?(?:\-)?\d+(?:\.(?:\d+)?)?', {'docstring':
        "Will match a full float, as well as a signed (and unsigned) integer"}
        ow                 = r'\s*', {'docstring':
        "\"Optional Whitechunk\""}

        @add_greedy_possessive
        @imply_pattern_is_cur
        def match_range(min, max, pattern=..., *, greedy=True, possessive=False, cur=...):
            """ Match between `min` and `max` sequences of `pattern` in the string. This also accepts `greedy` and `possessive` parameters
                Max can be an empty string to indicate no maximum
                `greedy` means it will try to match as many repititions as possible
                non-greedy will try to match as few repititions as possible
                `possessive` means it won't backtrack to try to find any repitions
                see your specific dialect docs for more help

                Args:
                    min (int): minimum number of instances of the pattern to match
                    max (int): maximum number of instances of the pattern to match
                    pattern: the pattern to match
                    greedy (bool): whether to match greedily (default: True)
                    possessive (bool): whether to match possessively (default: False)
            """
            raise_if_empty(pattern, 'match_range')

            s = cur
            if len(pattern):
                s += r'(?:' + pattern + r')'
            s += r'{' + str(min) + r',' + str(max) + r'}'
            if not greedy:
                s += r'?'
            if possessive:
                s += r'+'
            return s

        # Choices
        @add_greedy_possessive
        @imply_pattern_is_cur
        def optional(pattern=..., *, greedy=True, possessive=False, cur=...):
            """ Match `pattern` if it's there. This also accepts `greedy` and `possessive` parameters
                `greedy` means it will try to match as many repititions as possible
                non-greedy will try to match as few repititions as possible
                `possessive` means it won't backtrack to try to find any repitions
                see your specific dialect docs for more help

                Args:
                    pattern: the pattern to match
                    greedy (bool): whether to match greedily (default: True)
                    possessive (bool): whether to match possessively (default: False)
            """
            s = cur
            if len(pattern) > 1:
                s += fr'(?:{pattern})?'
            else:
                if len(pattern) == 1:
                    s += fr'{pattern}?'
            if not greedy:
                s += r'?'
            if possessive:
                s += r'+'
            return s

        def any_of(*patterns, chars=None, split=None, cur=...):
            """ Match any of the given `patterns`. Note that `patterns` can be multiple parameters,
                or a single string. Can also accept parameters chars and split. If char is set
                to True, then `patterns` must only be a single string, it interprets `patterns`
                as characters, and splits it up to find any of the chars in the string. If
                split is set to true, it forces the ?(...) regex syntax instead of the [...]
                syntax. It should act the same way, but your output regex will look different.
                By default, it just optimizes it for you.

                Args:
                    patterns: any of the patterns to match
                    chars (bool): whether to interpret patterns as characters (default: auto)
                    split (bool): whether to split patterns into characters (default: auto)
            """
            chars, patterns = _parse_any_of_params(*patterns, chars=chars, split=split)

            if chars:
                cur += r'['
                for i in patterns:
                    cur += i
                cur += r']'
            else:
                cur += r'(?:'
                for i in patterns:
                    cur += i
                    cur += '|'
                cur = cur[:-1]
                cur += r')'
            return cur

        def any_char_except(*chars, cur=...):
            """ This matches any char that is NOT in `chars`. `chars` can be multiple parameters,
                or a single string of chars to split.

                Args:
                    chars (str): any of the characters to match
            """

            # If it's just a string, split it up
            if len(chars) == 1 and len(chars[0]) > 1:
                chars = list(chars[0])

            cur += r'[^'
            for i in chars:
                cur += i
            cur += r']'
            return cur

        def any_between(char:str, and_char:str, cur=...):
            """ Match any char between `char` and `and_char`, using the ASCII table for reference

                Args:
                    char (str): the first character
                    and_char (str): the second character
            """
            raise_if_empty(char, 'any_between', 'char')
            raise_if_empty(and_char, 'any_between', 'and_char')
            return cur + r'[' + char + r'-' + and_char + r']'

        # TODO: the order of these parameters actually does matter, as sometime it needs
        # to match the first thing before it tries the 2nd.
        # @imply_pattern_is_cur
        def either(pattern, or_pattern=..., cur=...):
            """ Match either `pattern` or `or_pattern`. To choose between more than 2 things,
                you can either chain multiple `either` calls, or use `any_of`. Note that
                the order here matters: it first tries `pattern`, and if that doesn't
                match, then it tries `or_pattern`.

                Args:
                    pattern: a pattern to match
                    or_pattern: a pattern to match if the first one fails
            """
            # Manually handle implicit chains, instead of using imply_pattern_is_cur,
            # just because the order actually does matter
            if or_pattern is Ellipsis:
                return rf'(?:{cur}|{pattern})'
            return cur + rf'(?:{pattern}|{or_pattern})'

        # Amounts
        @imply_pattern_is_cur
        def match_max(pattern=..., *, cur=...):
            """ Match as many of `pattern` in the string as you can. This is equivelent to using the unary + operator.
            If `pattern` is not provided, it works on the previous regex pattern. That's not recommended for
            clarity's sake though

            Args:
                pattern: the pattern to match
            """
            return cur + r'(?:' + pattern + r')' + r'+'

        @imply_pattern_is_cur
        def match_num(num:int, pattern=..., *, cur=...):
            """ Match `num` amount of `pattern` in the string

            Args:
                num (int): the number of times to match the pattern
                pattern: the pattern to match
            """
            return cur + r'(?:' + pattern + r')' + r'{' + str(num) + r'}'

        @imply_pattern_is_cur
        def match_more_than(min:int, pattern=..., *, cur=...):
            """ Match more than `min` sequences of `pattern` in the string

            Args:
                min (int): the minimum number of times to match the pattern
                pattern: the pattern to match
            """
            return cur + r'(?:' + pattern + r')' + r'{' + str(int(min) + 1) + r',}'

        @imply_pattern_is_cur
        def match_at_least(min:int, pattern=..., *, cur=...):
            """ Match at least `min` sequences of `pattern` in the string

            Args:
                min (int): the minimum number of times to match the pattern
                pattern: the pattern to match
            """
            return cur + r'(?:' + pattern + r')' + r'{' + str(min) + r',}'

        @imply_pattern_is_cur
        def match_at_most(max:int, pattern=..., *, cur=...):
            """ Match at most `max` sequences of `pattern` in the string

            Args:
                max (int): the maximum number of times to match the pattern
                pattern: the pattern to match
            """
            return cur + r'(?:' + pattern + r')' + r'{0,' + str(max) + r'}'

        @add_greedy_possessive
        @imply_pattern_is_cur
        def at_least_one(pattern=..., *, greedy=True, possessive=False, cur=...):
            """ Match at least one of `pattern` in the string. This also accepts `greedy` and `possessive` parameters
                `greedy` means it will try to match as many repititions as possible
                non-greedy will try to match as few repititions as possible
                `possessive` means it won't backtrack to try to find any repitions
                see your specific dialect docs for more help

                Args:
                    pattern: the pattern to match
                    greedy (bool): whether to match greedily (default: True)
                    possessive (bool): whether to match possessively (default: False)
            """
            raise_if_empty(pattern, 'at_least_one')

            s = cur
            if len(pattern) > 1:
                s += fr'(?:{pattern})+'
            else:
                if len(pattern) == 1:
                    s += fr'{pattern}+'
            if not greedy:
                s += '?'
            if possessive:
                s += '+'
            return s

        @add_greedy_possessive
        @imply_pattern_is_cur
        def at_least_none(pattern=..., *, greedy=True, possessive=False, cur=...):
            """ Match 0 or more sequences of `pattern`. This also accepts `greedy` and `possessive` parameters
                `greedy` means it will try to match as many repititions as possible
                non-greedy will try to match as few repititions as possible
                `possessive` means it won't backtrack to try to find any repitions
                see your specific dialect docs for more help

                Args:
                    pattern: the pattern to match
                    greedy (bool): whether to match greedily (default: True)
                    possessive (bool): whether to match possessively (default: False)
            """
            raise_if_empty(pattern, 'at_least_none')

            s = cur
            if len(pattern) > 1:
                s += fr'(?:{pattern})*'
            else:
                if len(pattern) == 1:
                    s += fr'{pattern}*'
            if not greedy:
                s += '?'
            if possessive:
                s += '+'
            return s

    return _BaseMixin

# TODO: if group is called with 2 positional params, it raises an error. That error should be more helpful.
def GroupsMixin(*,
    named_group=lambda pattern, name, cur=...: f'{cur}(?P<{name}>{pattern})',
):
    """ A function which returns a class with the group and passive_group methods. Optionally
        specify how to handle named groups.
    """
    class _GroupsMixin:
        @imply_pattern_is_cur
        def group(pattern, *, name=None, cur=...):
            """ Causes `pattern` to be captured as a group. Specify `name` only as a keyword argument.
                Only useful when replacing regexs

                Args:
                    pattern: the pattern to match
                    name (str): the name of the group. Specifying this turns it into a named group,
                        instead of an unnamed group
            """
            raise_if_empty(pattern, 'group')
            if name is not None:
                if named_group is None:
                    raise ValueError(f'Named groups are not implemented in this dialect')
                raise_if_empty(name, 'group', 'name')
            return f'{cur}({pattern})' if name is None else named_group(pattern, name, cur=cur)

        @imply_pattern_is_cur
        def passive_group(pattern, *, cur=...):
            """ As all regexs in EZRegex capture passively, this is entirely useless. But if you really want to, here it is

                Args:
                    pattern: the pattern to match
            """
            raise_if_empty(pattern, 'passive_group')
            return f'{cur}(?:{pattern})'

    return _GroupsMixin

def AdvancedGroupsMixin(*,
        earlier_numbered_group=lambda num, cur=...: f'{cur}\\{num}',
        earlier_named_group=lambda name, cur=...: f'{cur}(?P={name})'
    ):
    class _AdvancedGroupsMixin:
        def earlier_group(num_or_name, cur=...):
            """ Matches whatever the group referenced by `num_or_name` matched earlier. Must be *after* a
                group which would match `num_or_name`

                Args:
                    num_or_name (int | str): either the number or name of the previous group
            """
            raise_if_empty(num_or_name, 'earlier_group', num_or_name)
            return earlier_numbered_group(num_or_name, cur=cur) \
                if isinstance(num_or_name, int) or num_or_name in digits \
                else earlier_named_group(num_or_name, cur=cur)

        def if_exists(num_or_name, does_pattern, doesnt_pattern=None, cur=...):
            """ Matches `does` if the group `num_or_name` exists, otherwise it matches `doesnt`

                Args:
                    num_or_name (int | str): either the number or name of the previous group
                    does_pattern: the pattern to match if the group exists
                    doesnt_pattern: the pattern to match if the group doesn't exist
            """
            raise_if_empty(num_or_name, 'if_exists', num_or_name)
            return f'{cur}(?({num_or_name}){does_pattern}{("|" + str(doesnt_pattern)) if doesnt_pattern is not None else ""})'

    return _AdvancedGroupsMixin

def AssertionsMixin():
    """ Also called "lookahead"/"lookbehind". Adds associated singleton members which use them """

    # TODO: enforce these only being at the end or beginning of a chain -- maybe
    class _AssertionsMixin:
        # TODO: this needs more tests
        def any_except(pattern, type='.*', cur=...):
            """ Matches anything other than `pattern`, which must be a single string
            or EZRegex chain. Optionally accepts the `type` parameter,
            which works like this: "Match any `type` other than `pattern`". For example,
            "match any word which is not foo".

            Args:
                pattern: the pattern to match
                type: the type of pattern to match
            """
            raise_if_empty(pattern, 'any_except')
            return cur + f'(?!{pattern}){type}'

        @imply_pattern_is_cur
        def if_proceded_by(pattern, *, cur=...):
            """ Matches the pattern if it has `pattern` coming after it. Can only be used once in a given pattern,
                as it only applies to the end

                Args:
                    pattern: the pattern to match
            """
            return fr'{cur}(?={pattern})'

        # TODO: this needs more tests
        def each(*patterns, cur=...):
            """ Matches if the next part of the string can match all of the given patterns.
                Like the + operator, but out of order.

                Args:
                    patterns: the patterns to match
            """
            patterns = list(patterns)
            last = patterns.pop()
            s = cur
            for i in patterns:
                s += fr'(?={i})'
            s += last
            return s

        # TODO: is this description correct?
        @imply_pattern_is_cur
        def if_not_proceded_by(pattern, *, cur=...):
            """ Matches the pattern if it does **not** have `pattern` coming after it. Can only be used once in
                a given pattern, as it only applies to the end

                Args:
                    pattern: the pattern to match
            """
            return fr'{cur}(?!{pattern})'

        @imply_pattern_is_cur
        def if_preceded_by(pattern, *, cur=...):
            """ Matches the pattern if it has `pattern` coming before it. Can only be used once in a given pattern,
                as it only applies to the beginning

                Args:
                    pattern: the pattern to match
            """
            return fr'(?<={pattern}){cur}'

        @imply_pattern_is_cur
        def if_not_preceded_by(pattern, *, cur=...):
            """ Matches the pattern if it does **not** have `pattern` coming before it. Can only be used once
                in a given pattern, as it only applies to the beginning

                Args:
                    pattern: the pattern to match
            """
            return fr'(?<!{pattern}){cur}'

        @imply_pattern_is_cur
        def if_enclosed_with(open, close=..., pattern=..., *, cur=...):
            """ Matches if the string has `open`, then `pattern`, then `close`, but only \"matches\"
                `pattern`. Just a convenience combination of if_proceded_by and if_preceded_by.

                Args:
                    open: the pattern to match before the stuff
                    close: the pattern to match after the stuff. If unspecified, assumes it's the same as `open`
                    pattern: the pattern to match
            """
            if close is Ellipsis:
                close = open
            return fr'((?<={open}){pattern}(?={close}))'

    return _AssertionsMixin

def AnchorsMixin(*, string=True, line=True, word_boundaries=True, word=True, string_end=r'\z'):
    """ String anchors, where string/line/word starts/ends. You can disable specific ones via parameters"""
    class _AnchorsMixin:
        if string:
            string_starts_with = lambda pattern='', cur=...: r'\A' + pattern + cur, {'docstring':
            """ Matches the string if it starts with `pattern`

                Args:
                    pattern: the pattern to match
            """}

            string_ends_with = lambda pattern='', cur=...: pattern + string_end + cur, {'docstring':
            """ Matches the string if it ends with `pattern`

                Args:
                    pattern: the pattern to match
            """}
            is_exactly = imply_pattern_is_cur(lambda pattern=..., cur=...: r"\A" + pattern + string_end), {'docstring':
            """ This matches the string if and only if the entire string is exactly equal to `pattern`

                Args:
                    pattern: the pattern to match
            """}

        if line:
            # Always use the multiline flag, so as to distinguish between start of a line vs start of the string
            line_starts_with = lambda pattern='', cur=...: r'^' + pattern + cur, {'flags':'m', 'docstring':
            """ Matches at a line if it starts with `pattern`

                Args:
                    pattern: the pattern to match
            """}
            line_ends_with = lambda pattern='', cur=...: cur + pattern + r'$', {'flags':'m', 'docstring':
            """ Matches at a line if it ends with `pattern`

                Args:
                    pattern: the pattern to match
            """}

        if word_boundaries:
            word_boundary = r'\b', {'docstring':
            """ Matches at a word boundary, i.e. the empty space between a word
                character and not a word character, or the end of a string
            """}
            not_word_boundary = r'\B', {'docstring':
            """ Matches at anything other than a word boundary, i.e. the empty
                space between a word character and not a word character, or the
                end of a string
            """}

        if word:
            word_starts_with = lambda pattern='', cur=...: r'\<' + pattern + cur, {'docstring':
            """ Matchs a word if it starts with `pattern`

                Args:
                    pattern: the pattern to match
            """}
            word_ends_with = lambda pattern='', cur=...: pattern + r'\>' + cur, {'docstring':
            """ Matchs a word if it ends with `pattern`

                Args:
                    pattern: the pattern to match
            """}

    return _AnchorsMixin

def ReplacementsMixin(*,
    named_group:None|Callable[[str,str],str]=lambda name, cur=...: f"{cur}${{{name}}}",
    numbered_group:None|Callable[[int,str],str]=lambda num, cur=...: f"{cur}${{{num}}}",
    entire_match:None|EZRegexFunc=None,
):
    # if entire_match isn't specified, most of the time the 0th number group is the same thing
    if entire_match is None:
        entire_match = partial(numbered_group, 0)

    def _rgroup(num_or_name, cur=...):
        """ Puts in its place the group specified, either by group number (for unnamed
            groups) or group name (for named groups). Named groups are typically also counted by
            number, check your specific dialect docs for details.
            Group 0 is handled specially by this function, so it calls for the entire match,
            even if 0 doesn't mean the entire match in your dialect.

            Args:
                num_or_name (int | str): the number or name of the group you want to insert here
        """
        raise_if_empty(num_or_name, 'rgroup', num_or_name)
        is_num = isinstance(num_or_name, int) or num_or_name in digits

        if not is_num and named_group is None:
            raise ValueError('named groups are not supported by this dialect')

        if is_num and numbered_group is None:
            raise ValueError('numbered groups are not supported by this dialect')

        if is_num and num_or_name in (0, '0'):
            return entire_match(cur=cur)

        return numbered_group(num_or_name, cur=cur) \
            if is_num \
            else named_group(num_or_name, cur=cur)

    class CustomFormatter(Formatter):
        def get_value(self, key, args, kwargs):
            return _rgroup(key, '')
    formatter = CustomFormatter()

    class _ReplacementsMixin:
        rliteral = lambda pattern, cur=...: cur + pattern, {'replacement': True, 'docstring':
        """ Exactly like literal, but for replacement regexs """}

        @EZRegex.exclude
        @classmethod
        def replace(cls, string, compile=True):
            """ Generates a valid regex replacement string, using Python f-string like syntax.

                Args:
                    string (str): the templated replacement string
                    compile (bool): whether to compile the string into an EZRegex subclass instance (default: True)

                Example:
                    ``` replace("named: {group}, numbered: {1}, entire: {0}") ```

                Like Python f-strings, use {{ and }} to specify { and }

                Set the `compile` parameter to False to have it return an EZRegex subclass instance instead of a string

                Note: 0 is handled specially by this function, so it calls for the entire match,
                    even if 0 doesn't mean the entire match in your dialect.

                There's a few of advantages to using this instead of just the regular regex replacement syntax:
                - It's consistent between dialects
                - It's closer to Python f-string syntax, which is cleaner and more familiar
                - It handles numbered, named, and entire replacement types the same
            """
            string = formatter.format(string)
            return string if compile else cls([lambda cur=...: cur + string], replacement=True)

        rgroup = _rgroup, {'replacement': True}
        replace_entire = entire_match, {'replacement': True, 'docstring':
        "Puts in its place the entire match"}

    return _ReplacementsMixin

def AdvancedReplacementsMixin():
    class _AdvancedReplacementsMixin:
        entire_string = '$_', {'replacement': True}
        string_before_match = '$`', {'replacement': True}
        string_after_match = "$'", {'replacement': True}

    return _AdvancedReplacementsMixin



import pytest

import ezregex as ez
from ezregex import *
from ezregex import EZRegex, python


def test_no_empty_strings():
    with pytest.raises(ValueError):
        str(group('', name='i') + optional(digit))

    with pytest.raises(ValueError):
        str(match_max(group('')))

    with pytest.raises(ValueError):
        str(amt(2, ''))

    with pytest.raises(ValueError):
        str(group('abc', name=''))


def test_any_of():
    # """ _any_of_func """
    assert ez.anyof('aiLmsux', split=True, chars=True)._compile(add_flags=False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{ez.anyof('aiLmsux', split=True, chars=True)._compile(add_flags=False)}'"
    # assert er.anyof('aiLmsux', split=False, chars=True)._compile(add_flags=False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=True)._compile(add_flags=False)}'"
    assert ez.anyof('aiLmsux', split=None, chars=True)._compile(add_flags=False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{ez.anyof('aiLmsux', split=None, chars=True)._compile(add_flags=False)}'"
    assert ez.anyof('aiLmsux', split=True, chars=False)._compile(add_flags=False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{ez.anyof('aiLmsux', split=True, chars=False)._compile(add_flags=False)}'"
    # assert er.anyof('aiLmsux', split=False, chars=False)._compile(add_flags=False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=False)._compile(add_flags=False)}'"
    assert ez.anyof('aiLmsux', split=None, chars=False)._compile(add_flags=False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{ez.anyof('aiLmsux', split=None, chars=False)._compile(add_flags=False)}'"
    assert ez.anyof('aiLmsux', split=True, chars=None)._compile(add_flags=False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{ez.anyof('aiLmsux', split=True, chars=None)._compile(add_flags=False)}'"
    # assert er.anyof('aiLmsux', split=False, chars=None)._compile(add_flags=False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=None)._compile(add_flags=False)}'"
    assert ez.anyof('aiLmsux', split=None, chars=None)._compile(add_flags=False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{ez.anyof('aiLmsux', split=None, chars=None)._compile(add_flags=False)}'"

    # assert er.anyof(*list('aiLmsux'), split=True, chars=True)._compile(add_flags=False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=True)._compile(add_flags=False)}'"
    assert ez.anyof(*list('aiLmsux'), split=False, chars=True)._compile(add_flags=False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{ez.anyof(*list('aiLmsux'), split=False, chars=True)._compile(add_flags=False)}'"
    assert ez.anyof(*list('aiLmsux'), split=None, chars=True)._compile(add_flags=False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{ez.anyof(*list('aiLmsux'), split=None, chars=True)._compile(add_flags=False)}'"
    # assert er.anyof(*list('aiLmsux'), split=True, chars=False)._compile(add_flags=False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=False)._compile(add_flags=False)}'"
    assert ez.anyof(*list('aiLmsux'), split=False, chars=False)._compile(add_flags=False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{ez.anyof(*list('aiLmsux'), split=False, chars=False)._compile(add_flags=False)}'"
    assert ez.anyof(*list('aiLmsux'), split=None, chars=False)._compile(add_flags=False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{ez.anyof(*list('aiLmsux'), split=None, chars=False)._compile(add_flags=False)}'"
    # assert er.anyof(*list('aiLmsux'), split=True, chars=None)._compile(add_flags=False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=None)._compile(add_flags=False)}'"
    assert ez.anyof(*list('aiLmsux'), split=False, chars=None)._compile(add_flags=False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{ez.anyof(*list('aiLmsux'), split=False, chars=None)._compile(add_flags=False)}'"
    assert ez.anyof(*list('aiLmsux'), split=None, chars=None)._compile(add_flags=False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{ez.anyof(*list('aiLmsux'), split=None, chars=None)._compile(add_flags=False)}'"

    assert ez.anyof('foo', word, 'bar', digit)._compile(add_flags=False) == r'(?:foo|\w+|bar|\d)'


def test_any_char_except():
    # """ _any_char_except_func """
    assert ez.any_char_except('abcd')._compile(add_flags=False) == '[^abcd]'
    assert ez.any_char_except(*list('abcd'))._compile(add_flags=False) == '[^abcd]'


def test_misc():
    """ Just threw a bunch of stuff in here for now, I need to move it elsewhere eventually """
    a = word + ow
    # b = stuff + UNICODE
    c = options('ignore_case') + '9'
    assert a + c == word + ow + options('ignore_case') + '9', f"{a + c} != {word + ow + options('ignore_case') + '9'}"

    a = str(PythonEZRegex([lambda cur=...: cur + r'\s+']))
    with pytest.raises(TypeError):
        b = str(PythonEZRegex(raw(r'\s+')))
    c = r'\s+'
    d = str(raw(r'\s+'))
    # e = str(whitespace + matchMax)
    assert a == c == d, f'\na: {a}\n c: {c}\n d: {d}\n e: {e}'
    # assert (word + ow + anything + ':').test('word    d:', show=False)
    # assert not (word + ow + anything + ':').test('word', show=False)
    assert 'word    d:' in (word + ow + anything + ':')

    test = word + chunk
    test += word
    assert str(test) == str(word + chunk + word), f"{str(test)} != {str(word + chunk + word)}"
    assert test == word + chunk + word
    assert either('(' + word + ')', '.') == either(ez.literal('(') + word() + ez.literal(')'), '.'), f"{either('(' + word + ')', '.')} != {either(ez.literal('(') + word() + ez.literal(')'), '.')}"
    assert str(ez.ifFollowedBy(word)) == r'(?=\w+)'

    #TODO: assert (word + ow + anything + ':') in 'word    d:'
    #TODO: assert (word + ow + anything + ':') not in 'word'
    assert str(word) == r'\w+', f'{word}'


def test_options():
    assert options().str() == ''
    with pytest.raises(ValueError):
        options('notaflag')
    with pytest.raises(ValueError):
        options(notaflag=True)
    with pytest.raises(ValueError):
        options(notaflag=False)
    assert str(word + options('IGNORE_CASE')) == r'(?i)\w+'
    assert str(word + options('ignore_case')) == r'(?i)\w+'
    assert str(word + options(ignore_case=True)) == r'(?i)\w+'
    assert str(word + options(IGNORE_CASE=True)) == r'(?i)\w+'
    # with pytest.raises(ValueError):
    # I don't *think* this should raise an error
    anyof('a', 'b', 'c') + options('ignore_case') + options(IGNORE_CASE=False)


def test_docstrings():
    shouldbe = """
Documentation:
    https://docs.python.org/3/library/re.html#flags

Usage:
    word + options(ignore_case=True)
    word + options('ignore_case')
    word + options('ignore_case', 'multiline')
    word + options('ignore_case', multiline=True)

Args:
    ascii:
        Make matching words, word boundaries, digits, and whitespace perform ASCII-only matching instead of full Unicode matching (which is default). This is only meaningful for Unicode (str) patterns, and is ignored for bytes patterns
    ignore_case:
        Perform case-insensitive matching, including expressions that explicitly use uppercase members. Full Unicode matching (such as Ü matching ü) also works unless the ASCII flag is used to disable non-ASCII matches. The current locale does not change the effect of this flag unless the LOCALE flag is also used
    single_line:
        Not recommended. Makes the '.' special character match any character at all, including a newline. It's recommended you simply use literally_anything instead
    locale:
        Try not to use this, and rely on unicode matching instead
    multiline:
        Not recommended. Makes the '^' and '$' special characters match the start and end of lines, instead of the start and end of the string. This is automatically inserted when using line_start and line_end, you shouldn't need to add it manually
    unicode:
        Match using the full unicode standard, instead of just ASCII. Enabled by default, and therefore redundant
""".strip()
    def make_whitespace_visible(s):
        return s.replace(' ', '•').replace('\n', '¶\n').replace('\t', '→→→→')
    # print(make_whitespace_visible(options.__doc__).strip())
    # print(make_whitespace_visible(shouldbe))
    assert options.__doc__.strip() == shouldbe

    # test mixin members
    # assert number.__doc__ == "Matches multiple digits next to each other. Does not match negatives or decimals"
    # test mixin methods
    # assert any_of.__doc__ == """ Match any of the given `inputs`. Note that `inputs` can be multiple parameters,
    #             or a single string. Can also accept parameters chars and split. If char is set
    #             to True, then `inputs` must only be a single string, it interprets `inputs`
    #             as characters, and splits it up to find any of the chars in the string. If
    #             split is set to true, it forces the ?(...) regex syntax instead of the [...]
    #             syntax. It should act the same way, but your output regex will look different.
    #             By default, it just optimizes it for you.
    #         """
    # # test mixin lambdas
    # assert unicode.__doc__ == "Matches a unicode character by name"

    # # test mixin decorated methods
    # assert match_range.__doc__ == """ Match between `min` and `max` sequences of `input` in the string. This also accepts `greedy` and `possessive` parameters
    #             Max can be an empty string to indicate no maximum
    #             `greedy` means it will try to match as many repititions as possible
    #             non-greedy will try to match as few repititions as possible
    #             `possessive` means it won't backtrack to try to find any repitions
    #             see https://docs.python.org/3/library/re.html for more help
    #         """

    # # test psuedonyms
    # assert is_exactly.__doc__ == exactly.__doc__ == "This matches the string if and only if the entire string is exactly equal to `input`"

    # assert word.digit.__doc__ == None


from typing import Any, Callable, Tuple, Dict, Unpack, List, Set
from types import FunctionType

try:
    from mypy_extensions import DefaultNamedArg, VarArg
    type EZRegexFunc = Callable[[VarArg, DefaultNamedArg("cur", str)], str]
except ImportError:
    type EZRegexFunc = Callable[[..., str], str]

type EZRegexType = 'EZRegex'
type EZRegexDefinition = str|EZRegexFunc|Tuple[str|EZRegexFunc, Dict[str, Any]]
type EZRegexOther = str|EZRegexType|int
type EZRegexParam = str|EZRegexType|int|bool|None



def inject_parts(cls):
    rtn = {i: getattr(cls, i) for i in cls.parts()}
    # It's technically a function, so it's handled differently
    if hasattr(cls, 'replace'):
        rtn['replace'] = cls.replace
    return rtn



""" Support for the PCRE2 dialect of regular expressions"""
__version__ = '1.1.0'

from ..EZRegex import EZRegex
from ..mixins import (AdvancedGroupsMixin, AdvancedReplacementsMixin, BaseMixin, AssertionsMixin, GroupsMixin, AnchorsMixin, ReplacementsMixin)
from ..flag_docs import common_flag_docs

class PCRE2EZRegex(
    BaseMixin(allow_greedy=True, allow_possessive=True),
    AssertionsMixin(),
    GroupsMixin(),
    AdvancedGroupsMixin(),
    AnchorsMixin(string=False),
    ReplacementsMixin(entire_match='$&'),
    AdvancedReplacementsMixin(),
    EZRegex,

    escape_chars=b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f',
    repl_escape_chars=b'$',
    flags={
        'global': 'g',
        'multiline': 'm',
        'ignore_case': 'i',
        'verbose': 'x',
        'single_line': 's',
        #  'unicode': 'u',
        #  'extra': 'X',
        'lazy': 'U',
        #  'anchor': 'A',
        'duplicate_groups': 'J',
        'noncapturing': 'n'
    },
    flags_docs_map={
        **common_flag_docs,
        'noncapturing': '''Not recomendded. Don't capture with any groups. Instead, simply don't use any groups''',
        'extra': '''Not recomendded. Any character following a \\ that is not a valid meta sequence \
will be faulted and raise an error. \\O, for example, will cause an \
error, and it will not match.'''
    }
):
    """
    Official docs:
    https://www.pcre.org/current/doc/html/pcre2syntax.html
    https://www.pcre.org/current/doc/html/pcre2pattern.html
    """
    def _final_func(self, s:str) -> str:
        if self.replacement:
            # This is how you escape a $ in a replacement string
            return s.replace(r'\$', '$$')
        return s


""" Support for the JavaScript dialect of regular expressions"""
__version__ = '1.1.0'

from .. import EZRegex
from ..mixins import (AdvancedGroupsMixin, BaseMixin, AssertionsMixin, GroupsMixin, AnchorsMixin, ReplacementsMixin)
from ..flag_docs import common_flag_docs

class JavascriptEZRegex(
    BaseMixin(allow_greedy=True, allow_possessive=True),
    AssertionsMixin(),
    GroupsMixin(named_group=lambda input, name, cur=...: f'{cur}(?<{name}>{input})'),
    AdvancedGroupsMixin(
        earlier_numbered_group=lambda num, cur=...: fr'{cur}\{num}',
        earlier_named_group=lambda name, cur=...: fr'{cur}\k<{name}>'
    ),
    AnchorsMixin(),
    ReplacementsMixin(
        named_group=lambda name, cur=...: fr'{cur}$<{name}>',
        numbered_group=lambda num, cur=...: fr'{cur}${num}',
        entire_match='$&',
    ),
    EZRegex,

    escape_chars=b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f/',
    flags={
        'global': 'g',
        'has_indicies': 'd',
        'multiline': 'm',
        'ignore_case': 'i',
        'single_line': 's',
        'unicode': 'u',
        'sticky': 'y'
    },
    flags_docs_map={
        **common_flag_docs,
        'has_indicies': '''Generate indices for substring matches''',
        'sticky': '''The pattern is forced to become anchored at the start of the search or at the position of the last successful match, equivalent to a \\G''',
    },
    flags_docs_link='https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_expressions#advanced_searching_with_flags',

    variables={
        '_string_anchor_used': (False, lambda l, r: l or r),
    },
):
    """
    Official docs:
    https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_expressions
    """

    # We intentionally only add the // if we're adding flags
    # However, note that it will add // even if there are no flags
    def _flag_func(self, final):
        if self._string_anchor_used and 'm' in self.flags:
            raise ValueError('string_starts_with and string_ends_with don\'t work with the multiline flag')
        if self.replacement:
            return final
        return f'/{final}/{''.join(self.flags)}'

    # TODO: test this to make sure I'm right to exclude cur from the output
    is_exactly = lambda input, cur=...: r"^" + input + r"$"

    if_exists = None

    string_starts_with = lambda input='', cur=...: r'^' + input + cur, {'_string_anchor_used': True}
    string_ends_with   = lambda input='', cur=...: cur + input + r'$', {'_string_anchor_used': True}



from abc import ABC
from functools import partial
from inspect import signature
import logging
import re
from typing import Callable, Dict, List, Set

# Unpack is only available in 3.11+
try:
    from typing import Unpack
except ImportError:
    # This is wrong
    Unpack = dict

from .api import api
from .psuedonyms import psuedonyms, all_psuedonyms
# from ezregex.generate import generate_regex
from .invert import invert
from .types import EZRegexFunc, EZRegexType, EZRegexDefinition, EZRegexOther, EZRegexParam

# TODO: Seperate EZRegex into a "bytes" mode vs "string" mode
# TODO: consider changing add_flags to "outer" or "end" or something
# TODO: a lot of the raised ValueErrors should probably a custom Exception. Something like UnimplementedDialect or something

# Because we have a method named str, some of the typing thinks we mean it instead of the type
_str = str

# TODO: this should be a custom error
def _raise_replacement_interoperable_error():
    raise TypeError('Dont mix replacement EZRegexs with non-replacement EZRegexs')

# These are here because having it be a class member as part of the parent and child classes
# seems to cause problems
_initial_variables = {
    # Cast to sets so it can accept strings
    'flags'             : (set(), lambda l, r: set(l) | set(r)),
    # None means it works with both. If the instance is a str (like rgroup(1) + 'foo'), it can be either
    'replacement'       : (lambda obj: None if isinstance(obj, str) else False, lambda l, r: l if l == r or l is None or r is None else _raise_replacement_interoperable_error()),
    # TODO: I think I removed this, try removing it and see if it breaks anything
    '_options_specified': (False, lambda l, r: l or r),
}
""" These propagate through the EZRegex chain, in ways defined by the lambda.
    The first value is either the default, or  a function which takes the current object
        and returns it's default
    The second is a function used to combine the values.
    The first value, if a callable, takes 1 argument, the object in question (which may not be an EZRegex instance!),
        and returns the default value for it
    The second value takes 2 arguments, the left and right values, and returns the combined value.
"""

def _to_camel_case(s):
    return ''.join((word.capitalize() if cnt else word) for cnt, word in enumerate(s.split('_')))

class EZRegex(ABC):
    """ Represents a part of regular expression syntax.
        If you're seeing this, and looking for documentation, go here: https://ezregex.readthedocs.io/en/latest/
    """

    # then and invert are added dynamically (for convenience), and they don't start with __,
    # so they have to be excluded here
    _exclusions = ['_escape_chars', '_repl_escape_chars', '_variables', 'invert', 'then']
    "Excluded methods"
    _added_vars = {}
    "A dict of {method name: dict} to manually add variables to methods"

    lazy_check_params = False
    "If True, parameters passed to singleton methods will not be checked until the regex is compiled"

    _deleted: set[_str] = set()
    "Because we can't actually delete inherited class members, we keep track of deleted members here"

    # For linting's sake
    flags: set[_str]
    replacement: bool
    docstring: _str
    _options_specified: bool
    _func_list: list[EZRegexFunc]
    _is_raw: bool

    @classmethod
    def exclude(cls:type, method:Callable):
        """ Exclude a method from being instantiated as a singleton member """
        cls._exclusions.append(method.__name__)
        return method

    # I thiiiiink this will work?
    @classmethod
    def add_vars(cls, method:Callable):
        """ Add variables to a method, similar to the `singleton = "abc", {"flags": "m"}` syntax """
        def inner(**kwargs):
            cls._added_vars[method.__name__] = kwargs
            return method
        return inner

    # TODO: this funciton is a stub
    @staticmethod
    def _get_variable_docstring(type_:type, definition:EZRegexDefinition):
        """ Get the docstring for a variable. The proper way to do this would be
            via ast parsing, but this works for now
            TODO: See _pep224_docstrings in https://github.com/pdoc3/pdoc/blob/af49da6151d6e5c2b575c105693e61af87808bda/pdoc/__init__.py#L255
        """
        # If given the line after it's definition, this would get the docstring (the named doc group)
        # quote = either('\'', '"')
        # optional(whitechunk) + group(either(amt(3, quote), quote), name='y') + group(match_max(literallyAnything), name='doc') + earlier_group('y')
        # docstring_regex = r"""(?:\s+)?(?P<y>(?:(?:(?:'|")){3}|(?:'|")))(?P<doc>(?:(?:.|\n))+)(?P=y)"""
        return

    @staticmethod
    def _interpret_definition(type_:type, definition:EZRegexDefinition|list):
        """ Interpret a definition into an instantiated EZRegex subclass object of type `type_`"""

        if isinstance(definition, str):
            return type_([lambda cur=...: cur + definition], docstring=EZRegex._get_variable_docstring(type_, definition))

        elif isinstance(definition, tuple):
            assert len(definition) == 2, f'Definition {definition} is not a tuple of length 2'
            assert isinstance(definition[1], dict), f'Definition {definition} is a tuple of 2, but the second element is not a dictionary'
            if type(definition[0]) is str:
                assert all(k in (*type_._variables, 'docstring', '_is_raw') for k in definition[1].keys()), f'Definition {definition} is a tuple of 2, but not all of the keys are `docstring`, `_is_raw`, or in the dialect\'s variables (avilable variables: {type_._variables})'
                return type_([lambda cur=...: cur + definition[0]], **definition[1])
            elif callable(definition[0]):
                d = definition[1].copy()
                try:
                    if definition[0].__name__ != '<lambda>':
                        d['docstring'] = definition[0].__doc__
                # There's a partial in one, which is callable, but doesn't have a __name__ attribute
                except AttributeError:
                    pass
                return type_([definition[0]], **d)
            else:
                raise ValueError(f'Definition {definition} is a tuple of 2, but the first element is not a string or callable')

        elif callable(definition):
            # Make sure the callable has the right signature
            sig = signature(definition)
            assert 'cur' in sig.parameters, f'Definition {definition} does not have cur as a keyword parameter'
            assert sig.parameters['cur'].default == ..., f'Definition {definition} does not have cur as a keyword parameter with default value of ...'
            return type_([definition], docstring=definition.__doc__, **type_._added_vars.get(definition.__name__, {}))

        elif isinstance(definition, list):
            return type_(definition)

    @staticmethod
    def _parse_options_params(flag_map, *args, normalize_case=True, **kwargs):
        if normalize_case:
            for key, value in flag_map.copy().items():
                flag_map[key.upper()] = flag_map[key.lower()] = value

        flags = set()
        for arg in args:
            if arg in flag_map:
                flags.add(flag_map[arg])
            else:
                raise ValueError(f'Unknown flag: {arg}. Available flags: {tuple(flag_map.keys())}')

        for key, value in kwargs.items():
            if key in flag_map:
                if value:
                    flags.add(flag_map[key])
                else:
                    try:
                        flags.remove(flag_map[key])
                    except KeyError:
                        pass
            else:
                raise ValueError(f'Unknown flag: {key}. Available flags: {tuple(flag_map.keys())}')

        return ''.join(flags)

    @staticmethod
    def _generate_options_from_flags(type_:type, flag_map, normalize_case=True, docs_map={}, docs_link=''):
        # This is a function, not an EZRegex subclass, by intention, even though it's used like one
        def options(*args, **kwargs):
            flags = EZRegex._parse_options_params(flag_map, *args, normalize_case=normalize_case, **kwargs)
            return type_([lambda cur=...: cur], flags=flags, options_specified=True)

        docs = ''
        if docs_link:
            docs += f"Documentation:\n    {docs_link}\n\n"
        docs += '''Usage:
    word + options(ignore_case=True)
    word + options('ignore_case')
    word + options('ignore_case', 'multiline')
    word + options('ignore_case', multiline=True)

Args:
'''
        _docs = {k.lower(): v for k, v in docs_map.items()}
        for flag in flag_map.keys():
            try:
                docs += f"    {flag.lower()}:\n        {_docs[flag.lower()]}\n"
            except KeyError:
                docs += f"    {flag.lower()}\n"
        options.__doc__ = docs

        return options

    def __init_subclass__(
        cls:type,
        escape_chars:bytes,
        flags:dict[str, str],
        repl_escape_chars:bytes=b'',
        flags_docs_map:dict[str, str]={},
        flags_docs_link:str='',
        variables:Unpack[Dict[str, EZRegexDefinition]]={},
        **kwargs
    ):
        # Validate & set escape_chars
        assert isinstance(escape_chars, bytes), f'Escape chars {escape_chars} is not bytes'
        assert isinstance(repl_escape_chars, bytes), f'Replacement escape chars {repl_escape_chars} is not bytes'
        cls._escape_chars = escape_chars
        cls._repl_escape_chars = repl_escape_chars

        # Validate & set variables
        for v in variables.values():
            assert isinstance(v, tuple), f'Value {v} is not a tuple'
            assert len(v) == 2, f'Value {v} is not a tuple of length 2'
            assert callable(v[1]), f'Value {v} is a tuple of 2, but the second element is not callable'
            # Make sure the callable has the right signature
            sig = signature(v[1])
            assert len(sig.parameters) == 2, f'Value {v} is a tuple of 2, but the second element is not a callable with 2 parameters'
        cls._variables = variables | _initial_variables

        # Instantiate members & methods
        # psuedonyms don't exist yet, so we're fine
        for name in cls.parts():
            value = getattr(cls, name)
            if value is None:
                # Apparently del and delattr(), when applied to class members, only
                # delete from the current class, if you try to delete a class
                # member which is inherited, it will still exist.
                # This is actually good, because we don't want to delete from a Mixin,
                # then later try to use that Mixin for a different dialect and have
                # it be gone.
                # Instead we keep track of what's "deleted" and prevent it from being
                # accessed via __get_attribute__
                cls._deleted.add(name)
            else:
                to = EZRegex._interpret_definition(cls, value)
                setattr(cls, name, to)
                setattr(cls, _to_camel_case(name), to)

        # Validate flag params, and generate the options function
        assert isinstance(flags, dict), f'Flags {flags} is not a dictionary'
        assert isinstance(flags_docs_map, dict), f'Flags docs map {flags_docs_map} is not a dictionary'
        assert isinstance(flags_docs_link, str), f'Flags docs link {flags_docs_link} is not a string'
        cls.options = cls._generate_options_from_flags(cls, flags, True, flags_docs_map, flags_docs_link)

        # For the sake of brevity, these are here. No different than being defined below
        # There's no particular reason I put these here instead of in __init__()
        cls.__imul__ = cls.__mul__
        cls.__iadd__ = cls.__add__
        # The shift operators just shadow the add operators
        # I don't think right and left shifts should be any different, right?
        cls.__lshift__ = cls.__rshift__ = cls.__add__
        cls.__rlshift__ = cls.__rrshift__ = cls.__radd__
        cls.__ilshift__ = cls.__irshift__ = cls.__iadd__
        # I just think this is a good psuedonym
        cls.then = cls.append

        cls.invert = cls.__invert__ = cls.__reversed__ = cls.inverse

        # Add all the psuedonymns
        for name, ps in psuedonyms.items():
            if hasattr(cls, name):
                value = getattr(cls, name)
                for p in ps:
                    setattr(cls, p, value)
                    # also add camelCase versions of the psuedonyms
                    setattr(cls, _to_camel_case(p), value)

        # Make the subclass immutable
        cls.__setattr__ = cls._raise_immutibility
        cls.__delattr__ = cls._raise_immutibility
        cls.__delete__ = cls._raise_immutibility
        cls.__set__ = cls._raise_immutibility

        return super().__init_subclass__(**kwargs)

    def __init__(self, func_list:list[EZRegexFunc]=[], _is_raw=False, docstring:str=None, **variable_values):
        # Use the defaults, which can get overriden if we're given variable values (i.e. by _combine())
        self.__dict__.update({k: (v[0](self) if callable(v[0]) else v[0]) for k, v in self._variables.items()})
        self.__dict__.update(variable_values)
        self.__dict__['_is_raw'] = _is_raw
        self.__dict__['docstring'] = docstring

        # Now that we're instantiated, we're immutable
        self.__dict__['_func_list'] = func_list

    def _combine(self, other:EZRegexOther, cls:type, add_to_end:bool=True, compile:bool=False, propogate_raw:bool=False):
        if isinstance(other, EZRegex) and not isinstance(other, type(self)):
            raise ValueError('Cannot combine EZRegex objects of different dialects')

        if add_to_end:
            func_list = self._func_list + self._sanitize_other(other, compile=compile)
            # func_list = self._func_list + [lambda cur=...: cur + self._sanitize_input(other)]
        else:
            func_list = self._sanitize_other(other, compile=compile) + self._func_list
            # func_list = [lambda cur=...: self._sanitize_input(other) + cur] + self._func_list
        return cls(
            func_list,
            # Use the combination functions to combine & propagate the variables
            **{
                # combine_spec is (default_value/callable that returns default_value, lamdba l, r: ...)
                # Call the combine function
                var: combine_spec[1](
                    # With the left value
                    self.__dict__[var],
                    (
                        # And the right value, if it's an EZRegex instance (and of the same dialect, technically)
                        other.__dict__[var]
                        if isinstance(other, type(self))
                        # Otherwise, the right value is the default
                        else (
                            # If the default is a callable, call it with the right value (the one which isn't an EZRegex instance)
                            combine_spec[0](other)
                            if callable(combine_spec[0])
                            # Otherwise, just use the default value
                            else combine_spec[0]
                        )
                    )
                )
                for var, combine_spec in
                self._variables.items()
            },
            # `other` is the right most operand, so it should be the raw value when it gets called.
            # If we have raw on the left side, then they're trying to do something like digit.raw.word,
            # which will get called out when they try to compile, so we can ignore it
            _is_raw=other._is_raw if propogate_raw else False,
        )

    def _sanitize_param(self, i:EZRegexParam, add_flags:bool=False):
        """ Sanitize things that are passed as parameters to a singleton member. """
        # If it's another chain, compile it
        if isinstance(i, EZRegex):
            return i._compile(add_flags=add_flags)
        # It's a string (so we need to escape it)
        # If this is a replacement string, it will automatically escape based on _repl_escape_chars
        elif isinstance(i, str):
            return self._escape(i)
        # Some parameters can be None or bools
        elif isinstance(i, bool) or i is None:
            return i
        # If parameters are ints, convert them to strings, and just know that when writing mixins
        elif isinstance(i, int):
            return str(i)
        # It's something we don't know, try to cast it to a string anyway
        else:
            try:
                logging.warning(f"Type {type(i)} passed to EZRegex, auto-casting to a string. Special characters will will not be escaped.")
                return str(i)
            except Exception as e:
                raise ValueError(f'Incorrect type {type(i)} given to EZRegex parameter: Must be string or another EZRegex chain.') from e

    def _sanitize_other(self, other:EZRegexOther, compile:bool=True) -> List[Callable[[_str], _str]]:
        """ Sanitize things that are combined with the current chain (i.e. via +)
            Note that this is only called by _combine()
        """
        # So somethings we need to compile immediately, while others we don't. Consider:
        # digit + whitespace.opt
        # digit.whitespace.opt
        # In the first case, we need to compile whitespace.opt immediately, because they're
        # seperate chains
        # In the second case, if we compile immediately, we'll break left to right operator precedence rules

        # If it's another chain, add it's func_list to ours
        if isinstance(other, EZRegex):
            # Remember, flags are handled in _combine()
            return [lambda cur=...: cur + other._compile(add_flags=False)] if compile else other._func_list
        # It's a string (so we need to escape it)
        # If this is a replacement string, it will automatically escape based on _repl_escape_chars
        elif isinstance(other, str):
            return [lambda cur=...: cur + self._escape(other)]
        # Allow word + 1 -> "\w+1"
        elif isinstance(other, int):
            return [lambda cur=...: cur + str(other)]
        # It's something we don't know, try to cast it to a string anyway
        else:
            try:
                logging.warning(f"Attempting to combine type {type(other)} with EZRegex, auto-casting to a string. Special characters will will not be escaped.")
                return [lambda cur=...: cur + str(other)]
            except Exception as e:
                raise ValueError(f'Incorrect type {type(other)} given to EZRegex parameter: Must be string or another EZRegex chain.') from e

    def _escape(self, pattern:_str, replacement:bool=False):
        """ Available as a class method, so we can escape strings when defining singletons
            The user could use this, but I can't think of a reason they would want to.
        """
        # This function was modified from the one in /usr/lib64/python3.12/re/__init__.py line 255
        _special_chars_map = {i: '\\' + chr(i) for i in (self.repl_escape_chars if replacement else self.escape_chars)}
        return pattern.translate(_special_chars_map)

    def _raise_immutibility(self, *_args, **_kwargs):
        raise TypeError('EZRegex objects are immutable')

    def _compile(self, regex:_str='', add_flags:bool=True) -> _str:
        for func in self._func_list:
            regex = func(cur=regex)

        # Add the flags
        if add_flags:
            regex = self._flag_func(regex)

            # This is in the flags scope, because we only add flags at the very end
            regex = self._final_func(regex)
        return regex

    # Abstract methods
    def _flag_func(self, final:_str) -> _str:
        """ This function is called to add the flags to the regex. It gets called even if there are no flags """
        if self.flags and not self.replacement:
            return f'(?{''.join(self.flags)}){final}'
        return final

    def _final_func(self, compiled:_str) -> _str:
        """ Gets called just before returning the compiled string to the user. Useful for adding things like
            Javascript's slashes (i.e. /regex/)
        """
        return compiled

    def _escape(self, pattern:_str):
        return self.escape(pattern, self.replacement)

    @classmethod
    def escape(cls, pattern:_str, replacement:bool=False) -> _str:
        """ Available as a class method, so we can escape strings when defining singletons
            The user could use this, but I can't think of a reason they would want to.
        """
        # This function was modified from the one in /usr/lib64/python3.12/re/__init__.py line 255
        _special_chars_map = {i: '\\' + chr(i) for i in (cls._repl_escape_chars if replacement else cls._escape_chars)}
        return pattern.translate(_special_chars_map)

    # Regular functions
    @classmethod
    def parts(cls, include_psuedonyms=True, include_functions=True):#, include_replacement=True, _check_replacement=True):
        """ A utility function that lists all the names of all singleton methods in this dialect
            This excludes dunder methods, abstract methods, and any methods marked with @exclude

            Args:
                include_functions (bool): If False, it will exclude options and replace, both of which are
                    technically functions and not EZRegex instances.
                include_psuedonyms (bool): If False, it only return the canonical names
        """
        parent_members = dir(EZRegex)
        return [i
            for i in dir(cls)
            if (
                i not in parent_members and
                not i.startswith('__') and
                i not in cls._exclusions and
                i not in cls._deleted and
                # (include_psuedonyms or i not in all_psuedonyms) and
                (include_psuedonyms or i in psuedonyms) and
                (include_functions or i not in ('options', 'replace'))
                # Because we use this before singleton members are instantiated, getattr().replacement will fail
                # So in that case, we just skip it
                # (
                #     not _check_replacement or
                #     i == 'options' or
                #     include_replacement == (getattr(cls, i).replacement in (True, None))
                # )
            )
        ]

    def str(self):
        return self._compile()

    def copy(self, add_flags:bool=True):
        try:
            from clipboard import copy  # type: ignore
        except ImportError as err:
            raise ImportError(
                'Please install the clipboard module in order to auto copy ezregex expressions (i.e. pip install clipboard)'
            ) from err
        else:
            copy(self._compile(add_flags=add_flags))

    # TODO: this should use the api function instead
    def test(self, testString:str=None, show:bool=True, context:bool=True) -> bool:
        """ Tests the current regex expression to see if it's in @param testString.
            Returns the match objects (None if there was no match)
        """
        try:
            from rich import print as rprint
            from rich.panel import Panel
            from rich.text import Text
        except ImportError:
            raise ImportError("The rich library is required to use the EZRegex.test() method. Try running `pip install rich`")

        # json = self._matchJSON(testString=testString)
        json = api(self, test_string=testString)
        if not show:
            return bool(len(json['matches']))

        st = Text()  # String
        gt = Text()  # Groups (all the group-related text)
        defaultColor = 'bold'
        textColor = ''

        st.append("Testing expression", style=defaultColor)
        # Add the context line
        # st.append(f' (from {get_context(get_metadata(2), False, True, True).strip()})', style=defaultColor)
        st.append(':\n', style=defaultColor)

        # The expression we're testing
        st.append(f'\t{json["regex"]}\n', style=textColor)
        st.append("for matches in:\n\t", style=defaultColor)

        # Add the main string
        for color, background, part in json['parts']:
            st.append(part, style=color if background is None else f'{color} on {background}')
        st.append('\n')

        # Add all the matches and groups
        for m in json['matches']:
            gt.append('Match = "')
            for color, background, part in m['match']['parts']:
                gt.append(part, style=color if background is None else f'{color} on {background}')
            gt.append('" ')
            gt.append(f"({m['match']['start']}:{m['match']['end']})", style='italic bright_black')
            gt.append('\n')
            if len(m['unnamed groups']):
                gt.append('Unnamed Groups:\n')
            for id, group in m['unnamed groups'].items():
                gt.append(f'\t{id}: "')
                gt.append(group['string'], style=group['color'])
                gt.append('" ')
                gt.append(f"({group['start']}:{group['end']})", style='italic bright_black')
                gt.append('\n')
            if len(m['named groups']):
                gt.append('Named Groups:\n')
            for name, group in m['named groups'].items():
                gt.append(f'\t{name}: "')
                gt.append(group['string'], style=group['color'])
                gt.append('" ')
                gt.append(f"({group['start']}:{group['end']})", style='italic bright_black')
                gt.append('\n')
            gt.append('\n')

        # Assemble everything into a panel
        rprint(Panel(Text.assemble(*st, '\n', *gt),
            title='Testing Regex',
            subtitle=Text('Found\n', style='blue') if len(json['matches'])
                else Text('Not Found\n', style='red')))  #, border_style='dim green')

        return bool(len(json['matches']))

    def inverse(self, amt:int=1, **kwargs) -> str:
        """ "Inverts" the current Regex expression to give an example of a string it would match.
            Useful for debugging purposes. """
        return '\n'.join([invert(self._compile(), **kwargs) for _ in range(amt)])

    # TODO: think through this
    # def or_(self, input):
        # return self._base('either', self, input)

    # Named operator functions
    # TODO: ensure these have tests (there's a possibility these are backwards)
    def append(self, input:EZRegexOther) -> EZRegexType:
        return self._combine(input, type(self), compile=True)

    def prepend(self, input:EZRegexOther) -> EZRegexType:
        return self._combine(input, type(self), add_to_end=False, compile=True)

    # TODO: tests for this
    def concat(self, *args):
        """ Join multiple EZRegex singleton members together
        The following are all equivelent:
            Dialect.concat(a, b, c)
            Dialect.concat((a, b, c))
            Dialect.a.b.c
            a + b + c
            a.b.c
        """
        if len(args) == 1 and isinstance(args[0], (tuple, list)):
            args = args[0]

        if len(args) == 0:
            raise ValueError('concat requires at least one argument')

        chain = args[0]
        for i in args[1:]:
            chain += i

        return chain

    # Flag functions
    def set_flags(self, flags:_str|Set[_str]):
        """ Directly sets flags in an EZRegex instance
            NOTE: you most likely don't want this, you likely want .options()
            This sets flags directly, not the names of the flags
        """
        return type(self)(self._func_list, flags=set(flags))

    def add_flags(self, flags:_str|Set[_str]):
        """ Adds a flag to an EZRegex instance
            NOTE: you most likely don't want this, you likely want .options()
            This adds flags directly, not the names of the flags
        """
        return type(self)(self._func_list, flags=set(self.flags) | set(flags))

    def remove_flags(self, flags:_str|Set[_str]):
        """ Removes a flag from an EZRegex instance """
        return type(self)(self._func_list, flags=set(self.flags) - set(flags))

    # Magic Functions
    def __call__(self, *args:Unpack[List[EZRegexParam]], **kwargs:Unpack[Dict[_str, EZRegexParam]]):
        """ This should be called by the user to specify the specific parameters of this instance i.e. anyof('a', 'b') """
        # Alright, buckle up
        # Firstly, the following are true:
        # 1. singleton members are instances of their own class (EZRegex subclass)
        # 2. We have both __get__ and __call__ operators overloaded, which return modified instances/copies of the original instance
        # 3. __get__ and __call__ have the same precedence, so they get evaluated from left to right
        #   (https://docs.python.org/3/reference/expressions.html#evaluation-order)

        # Given these assumptions, consider the following:
        # word_char.match_num(4)
        # Because __get__ and __call__ have the same precedence, this is evaluated as:
        # (word_char.match_num)(4)
        # instead of
        # word_char.(match_num(4))
        # Which is unintuitive for my framework

        # To fix this, we pass along the entire current chain to the copy we create below,
        # and modify the last function instead of the only function.

        # This means that if you try to pass parameters to a singleton member that it can't accept,
        # it will throw an error, but if you call a chain with no parameters, there's really no
        # way (that I can think of right now) to tell if it's a chain or just at the end of a chain.
        # This means that you *can* do things like word_char() and word_char().match_num(4)(), which
        # I would like to discourage, but really have no way to prevent. It also won't necissarily
        # break anything, so it's fine.

        # Don't sanitize raw specifically
        if not self._is_raw:
            # Sanatize the positional arguments
            args = tuple(map(self._sanitize_param, args))

            # Sanatize the keyword arguments
            _kwargs = {}
            for key, val in kwargs.items():
                _kwargs[key] = self._sanitize_param(val)
        else:
            _kwargs = kwargs

        # Check the parameters are all valid
        if not type(self).lazy_check_params:
            sig = signature(self._func_list[-1])
            # This raises a TypeError if the arguments are invalid
            sig.bind(*args, cur=..., **_kwargs)

        # Create a copy, and pass the current variables to it, so they stay in the chain.
        # Also pass the current _func_list to the copy, and modify the last function to have
        # the parameters given to this function given to it (a partial call). This will
        # let it still require cur, while all the others get bound
        return type(self)(self._func_list[:-1] + [partial(self._func_list[-1], *args, **_kwargs)], **self.__dict__)

    def __get__(self, instance:EZRegexType|None, owner:type) -> EZRegexType:
        # lhs.rhs -> type(lhs).__dict__["rhs"].__get__(lhs, type(lhs))
        # self is rhs, instance is lhs, owner is type(lhs)
        # We're trying to access it as a class member. This is how chains are started
        if instance is None:
            return self
        # Remember, . (__get__) and () (__call__) have the same precedence, so they get evaluated from left to right
        # This means if we do digit.raw('foo'), it goes digit._combine(raw)('foo').
        return instance._combine(self, owner, compile=False, propogate_raw=True)

    def __eq__(self, other:EZRegexType) -> bool:
        """ NOTE: This will return False for equivelent EZRegex expressions of different dialects
            NOTE: This checks flags as well
        """
        if not isinstance(other, type(self)):
            return False
        if other.replacement != self.replacement and other.replacement is not None and self.replacement is not None:
            return False
        return other._compile() == self._compile()

    def __add__(self, other:EZRegexOther) -> EZRegexType:
        return self._combine(other, type(self), compile=True)

    def __radd__(self, other:EZRegexOther) -> EZRegexType:
        return self._combine(other, type(self), add_to_end=False)

    def __mul__(self, amt:int):
        if amt is Ellipsis:
            try:
                return type(self).at_least_none(self)
            except AttributeError:
                raise ValueError(f'at_least_none is not supported in {type(self).__name__}') from None

        rtn = self
        # This isn't optimal, but it's unlikely anyone will use this with large numbers
        for _ in range(amt-1):
            # This doesn't work
            # rtn += self
            # But this does??
            rtn = rtn + self
        return rtn

    def __rmul__(self, amt:int):
        return self.__mul__(amt)

    def __and__(self, other:EZRegexOther) -> EZRegexType:
        raise NotImplementedError

    def __rand__(self, other:EZRegexOther) -> EZRegexType:
        raise NotImplementedError

    def __pos__(self):
        try:
            return type(self).at_least_one(self)
        except AttributeError:
            raise ValueError(f'at_least_one is not supported in {type(self).__name__}') from None

    def __or__(self, other:EZRegexOther) -> EZRegexType:
        logging.warning('The or operator is unstable and likely to fail, if used more than twice. Use anyof() instead, for now.')
        try:
            return type(self).either(self, other)
        except AttributeError:
            raise ValueError(f'either is not supported in {type(self).__name__}') from None

        # return self._copy(f'(?:{self._compile(add_flags=False)}|{self._sanitizeInput(other)})', sanitize=False)

    def __ror__(self, other:EZRegexOther) -> EZRegexType:
        return other.__or__(self)

    def __xor__(self, other:EZRegexOther) -> EZRegexType:
        return NotImplementedError

    def __rxor__(self, other:EZRegexOther) -> EZRegexType:
        return NotImplementedError

    def __mod__(self, other:EZRegexOther) -> re.Match|None:
        """ I would prefer __rmod__(), but it doesn't work on strings, since __mod__() is already specified for string formmating. """
        # I don't need to check this, re will do it for me
        # if not isisntance(other, str):
            # raise TypeError(f"Can't search type {type(other)} ")
        return re.search(other, self._compile(add_flags=False))

    def __hash__(self) -> int:
        # TODO: should 2 different dialects who's regexs compile to the same thing have different hashs?
        if len(self._func_list) > 1:
            return hash(self._compile())
        # If we only have 1 function lined up, that means we haven't
        # been called at all. And that means we're one of the basic singletons,
        # because users aren't supposed to instantiate this class directly.
        # THAT means we can use this instance's pointer as a unique identifier.
        else:
            return hash(id(self))

    def __contains__(self, other:str) -> bool:
        # assert isinstance(other, str), "`in` statement can only be used with a string"
        return re.search(self._compile(), other) is not None

    def __rcontains__(self, other:str) -> bool:
        """ I guess this isn't a other? But it really should be. """
        # assert isinstance(other, str), "`in` statement can only be used with a string"
        return re.search(self._compile(), other) is not None

    def __getitem__(self, args:slice|tuple) -> EZRegexType:
        # digit[2, 3]    # (2, 3)
        # digit[2, ...]  # (2, Ellipsis)
        # digit[2, None] # (2, None)
        # digit[2, ]     # (2,)
        # digit[..., 3]  # (Ellipsis, 3)
        # digit[None, 3] # (None, 3)
        # digit[...:3]   # slice(Ellipsis, 3, None)
        # digit[None:3]  # slice(None, 3, None)
        # digit[:3]      # slice(None, 3, None)
        # digit[:]       # slice(None, None, None)
        # digit[2]       # 2

        # assert digit[2, 3] == match_range(2, 3, digit)
        # assert digit[2, ...] == digit[2,] == digit[2, None] == digit[2] == match_at_least(2, digit)
        # # assert digit[..., 2] == digit[0, 2] == digit[None, 2] == match_at_most(2, digit)
        # assert digit[...] == digit[0, ...] == digit[None] == at_least_0(digit)
        # assert digit[1, ...] == digit[1] == digit[1,] == digit[1, None] == at_least_1(digit)
        try:
            if type(args) is slice:
                # expr[...:end_expr] is equivalent to ZeroOrMore(expr, stop_on=end_expr)
                # assert digit[...:'foo'] == digit[None:'foo'] == digit[,'foo'] ==
                pass
            elif type(args) is not tuple or len(args) == 1:
                if type(args) is tuple:
                    args = args[0]
                if args is None or args is Ellipsis or args == 0:
                    return type(self).at_least_none(self)
                elif args == 1:
                    return type(self).at_least_one(self)
                else:
                    return type(self).match_at_least(args, self)
            else:
                start, end = args
                if start is None or start is Ellipsis:
                    return type(self).match_at_most(end, self)
                elif end is None or end is Ellipsis:
                    if start == 0:
                        return type(self).at_least_none(self)
                    elif start == 1:
                        return type(self).at_least_one(self)
                    else:
                        return type(self).match_at_least(start, self)
                else:
                    return type(self).match_range(start, end, self)
        except AttributeError as e:
            raise ValueError(f'That functionality is not implemented in {type(self).__name__}') from e
        except Exception as e:
            raise ValueError(f'Invalid arguments for __getitem__') from e

    def __str__(self):
        return self._compile()

    def __repr__(self):
        exclude_vars = ('_func_list', '_options_specified', '_is_raw', 'docstring', '_deleted')
        d = {k: v for k, v in self.__dict__.items() if k not in exclude_vars}
        return f'{type(self).__name__}({self._compile()}, {d})'

    def __getattribute__(self, name):
        if name in super().__getattribute__('_deleted'):
            raise NameError(f'name {name} is not defined')
        return super().__getattribute__(name)


# camelCase versions (of the original, and it's psuedonyms) are automatically added,
# so only include actually different psuedonyms
# If there is the same name on both sides (key and value), it causes complex issues
# The keys must be the cannonical names (i.e. the name that is used in the mixin or subclass)
# Don't forget that the values must be tuples, even if there is only one psuedonym
# TODO: more complete tests testing all of these

psuedonyms = {
    'match_max'          : ('repeat',),
    'match_at_most'      : ('at_most',),
    'match_num'          : ('match_amt', 'num', 'amt'),
    'match_range'        : ('between', 'match_between',),
    'match_more_than'    : ('match_greater_than', 'more_than',),
    'match_at_least'     : ('match_min', 'at_least',),
    'line_starts_with'   : ('line_start',),
    'string_starts_with' : ('string_start',),
    'line_ends_with'     : ('line_end',),
    'string_ends_with'   : ('string_end',),
    'chunk'              : ('stuff',),
    'whitechunk'         : ('whitespace', 'white_chunk', 'white_space'),
    'white_char'         : ('whitechar',),
    'anything'           : ('anychar', 'any_char', 'char',),
    'letter'             : ('alpha',),
    'letter_num'         : ('alphanum'  , 'alpha_num',),
    'any_between'        : ('num_between', 'amt_between',),
    'hex_digit'          : ('hex',),
    'new_line'           : ('newline',),
    'period'             : ('dot',),
    'any_of'             : ('anyof', 'one_of', 'oneof',),
    'any_char_except'    : ('any_except', 'anything_except',),
    'if_proceded_by'     : ('if_followed_by',),
    'if_not_proceded_by' : ('if_not_followed_by',),
    'if_enclosed_with'   : ('if_enclosed_by',),
    'earlier_group'      : ('same_as', 'same_as_group',),
    'is_exactly'         : ('exactly',),
    'optional'           : ('one_or_none', 'opt',),
    'at_least_one'       : ('one_or_more', 'at_least_1',),
    'at_least_none'      : ('none_or_more', 'at_least_0', 'any_amt', 'zero_or_more',),
    'signed'             : ('integer', 'signed_int', 'signed_integer',),
    'rgroup'             : ('replace_group',),
    'replace_entire'     : ('replace_all',),
    'either'             : ('or_',),
}

# TODO: this likely can be deleted, I don't think it's used anywhere anymore
all_psuedonyms = set()
for psuedonym in psuedonyms.values():
    all_psuedonyms.update(psuedonym)


import re
from timeit import repeat
from warnings import warn

import pytest

import ezregex as ez
from ezregex import *

ALL_DIALECTS = [ez.python.PythonEZRegex, ez.r.REZRegex, ez.javascript.JavascriptEZRegex, ez.pcre2.PCRE2EZRegex]

# TODO: tests to check how raw() interacts with replacement groups

def test_basic():
    assert literal('test').str() == 'test'

def test_eq():
    assert literal('test') + digit == 'test' + digit

def test_basic_concat():
    assert str(literal('test') + digit) == r'test\d'
    assert str('test' + digit) == r'test\d'

def test_access_dialect():
    assert type(literal('thing')) is PythonEZRegex

def test_psuedonyms():
    assert ez.matchMax(digit) == ez.match_max(digit)
    assert ez.python.matchMax(digit) == ez.python.match_max(digit)
    assert matchMax(digit) == match_max(digit)

def test_immutability():
    with pytest.raises(TypeError):
        digit.flags = 'ab'
    with pytest.raises(TypeError):
        digit.compile = lambda: ...
    with pytest.raises(TypeError):
        del digit.compile
    with pytest.raises(TypeError):
        del digit.flags

def test_test_method():
    return
    # ow = optional(whitechunk)
    params = ez.group(ez.atLeastNone(ez.ow + ez.word + ez.ow + ez.optional(',') + ez.ow))
    function = ez.word + ez.ow + '(' + params + ')'
    function.test('this should match func(param1, param2 ), foo(), and bar( foo,)')

    r = 'group 1' + ':' + ow + group('stuff') + ' | ' + 'group ' + number + ': ' + group('things') + ' | ' + 'named group "' + word + '": '  + named_group('foo', 'bar')
    s = 'random stuff! and then group 1: stuff | group 2: things | named group "foo": bar  \t oh and then more random stuff'
    r.test(s)

    s = 'word1 word2 word3'
    word.test(s)

    (word + whitechunk + group('func') + ':' + group(anyof('8', '7')), 'test').test()

    # This is actually accurate, if you think about it.
    # ifFollowedBy(word).test("literal(hllow) + isExactly('thing')")# fails in _matchJSON()

    ('(' + +(anything + optional(group(comma))) + ')').test()# -- empty groups print as None

    group(+group(number) + group(anyof('98'))).test('999')

def test_parameters_to_chains_lazy():
    EZRegex.lazy_check_params = True
    with pytest.raises(TypeError):
        digit + word(6).str()
    with pytest.raises(TypeError):
        word(6).str()
    with pytest.raises(TypeError):
        digit.word(6).str()

    with pytest.raises(TypeError):
        digit + word(pattern=6).str()
    with pytest.raises(TypeError):
        word(pattern=6).str()
    with pytest.raises(TypeError):
        digit.word(pattern=6).str()

    assert any_between('a', 'j').str() == r'[a-j]'
    with pytest.raises(TypeError):
        any_between('a', 'j', 'k').str()
    with pytest.raises(TypeError):
        any_between('a').str()
    # Lazy checking is less strict, it doesn't check types
    # with pytest.raises(TypeError):
    #     any_between(1.2, 3.4).str()
    # with pytest.raises(TypeError):
    #     any_between(1.2).str()
    # with pytest.raises(TypeError):
    #     any_between(1.2, 'a', 'j').str()
    # with pytest.raises(TypeError):
    #     any_between('a', 'j', 1.2).str()

    with pytest.raises(TypeError):
        digit + word(6).str()
    with pytest.raises(TypeError):
        word(6).str()
    with pytest.raises(TypeError):
        digit.word(6).str()

    with pytest.raises(TypeError):
        digit + word(pattern=6).str()
    with pytest.raises(TypeError):
        word(pattern=6).str()
    with pytest.raises(TypeError):
        digit.word(pattern=6).str()

    assert any_between('a', 'j').str() == r'[a-j]'
    assert any_between(1.2, 3.4).str()
    with pytest.raises(TypeError):
        any_between('a', 'j', 'k').str()
    with pytest.raises(TypeError):
        any_between('a').str()
    with pytest.raises(TypeError):
        any_between(1.2).str()
    with pytest.raises(TypeError):
        any_between(1.2, 'a', 'j').str()
    with pytest.raises(TypeError):
        any_between('a', 'j', 1.2).str()

    # There's nothing we can do about these, so just make sure they don't break anything
    # See the comment in EZRegex.py.__call__ for why this is
    assert digit().str() == r'\d'
    assert word().str() == r'\w+'
    assert digit.word().str() == r'\d\w+'
    assert digit().word().str() == r'\d\w+'
    assert match_amt(6, digit)().str() == r'(?:\d){6}'
    assert digit().str() == r'\d'
    assert word_char.amt(4).str() == r'(?:\w){4}'
    assert word_char.amt(4).group(name='word').literal(' ').str() == r'(?P<word>(?:\w){4})\ '

    # Reset, just in case
    EZRegex.lazy_check_params = False

def test_parameters_to_chains_eager():
    with pytest.raises(TypeError):
        digit + word(6)
    with pytest.raises(TypeError):
        word(6)
    with pytest.raises(TypeError):
        digit.word(6)

    with pytest.raises(TypeError):
        digit + word(pattern=6)
    with pytest.raises(TypeError):
        word(pattern=6)
    with pytest.raises(TypeError):
        digit.word(pattern=6)

    assert any_between('a', 'j').str() == r'[a-j]'
    with pytest.raises(TypeError):
        any_between('a', 'j', 'k')
    with pytest.raises(TypeError):
        any_between('a')
    with pytest.raises(TypeError):
        any_between(1.2)
    with pytest.raises(TypeError):
        any_between(1.2, 'a', 'j')
    with pytest.raises(TypeError):
        any_between('a', 'j', 1.2)

def test_empty_input():
    warn('This test works manually, but not in the testing env. Retest later')
    return
    with pytest.raises(ValueError):
        amt(3, '')

def test_re_shadow_funcs():
    s = r'\d(\w+)'
    string = 'timmy is 6years old'
    repl = replace('number {1}')

    def eq(a, b):
        if not ((a is None) == (b is None)) and a.span() == b.span() and a.groups() == b.groups():
            raise AssertionError(f'{a} != {b}')

    eq((digit + group(word)).search(string),      re.compile(s).search(string))
    eq((digit + group(word)).match(string),       re.compile(s).match(string))
    eq((digit + group(word)).fullmatch(string),   re.compile(s).fullmatch(string))
    eq((digit + group(word)).split(string),       re.compile(s).split(string))
    eq((digit + group(word)).findall(string),     re.compile(s).findall(string))
    eq((digit + group(word)).finditer(string),    re.compile(s).finditer(string))
    eq((digit + group(word)).sub(repl, string),   re.compile(s).sub(repl, string))
    eq((digit + group(word)).subn(repl, string),  re.compile(s).subn(repl, string))

    s = r'(?P<word>\w{4}) '
    string = 'this string should have 3 four letter words'
    repl = replace('4444 ({word}) ({0})')
    eq((word_char.amt(4).group(name='word').literal(' ')).search(string),      re.compile(s).search(string))
    eq((word_char.amt(4).group(name='word').literal(' ')).match(string),       re.compile(s).match(string))
    eq((word_char.amt(4).group(name='word').literal(' ')).fullmatch(string),   re.compile(s).fullmatch(string))
    eq((word_char.amt(4).group(name='word').literal(' ')).split(string),       re.compile(s).split(string))
    eq((word_char.amt(4).group(name='word').literal(' ')).findall(string),     re.compile(s).findall(string))
    eq((word_char.amt(4).group(name='word').literal(' ')).finditer(string),    re.compile(s).finditer(string))
    eq((word_char.amt(4).group(name='word').literal(' ')).sub(repl, string),   re.compile(s).sub(repl, string))
    eq((word_char.amt(4).group(name='word').literal(' ')).subn(repl, string),  re.compile(s).subn(repl, string))

def test_flag_methods():
    assert (digit + options('ascii')).flags == {'a'}
    with pytest.raises(TypeError):
        digit.flags = {'L'}
    assert digit.flags == set()
    a = digit.set_flags('asL')
    assert a.flags == {'a', 's', 'L'}
    assert (digit + options('ascii')).add_flags('L').flags == {'a', 'L'}
    assert (digit + options('ascii')).remove_flags('a').flags == set()
    assert (digit + options('ascii')).remove_flags('L').flags == {'a'}

def test_imply_input_is_cur():
    input = word
    min = 1
    max = 3

    assert amt(3, digit).str() == r'(?:\d){3}'
    assert digit.amt(3).str() == r'(?:\d){3}'
    assert digit.amt(3, 'a')
    assert digit.amt('a', 3)
    assert digit.amt('a', 'b')
    assert amt('a', 3)
    assert amt('a', 'b')

    try:
        assert digit.group() == group(digit) == digit.group
        assert digit.group(name='test') == group(digit, name='test')
        assert digit.if_not_preceded_by(input) == if_not_preceded_by(input) + digit
        assert digit.if_preceded_by(input) == if_preceded_by(input) + digit
        assert digit.if_not_proceded_by(input) == digit + if_not_proceded_by(input)
        assert digit.if_proceded_by(input) == digit + if_proceded_by(input)
        assert digit.if_enclosed_with('|') == if_enclosed_with('|', '|', digit) == if_enclosed_with('|', pattern=digit)
        assert digit.at_least(min) == at_least(min, digit)
        assert digit.more_than(min) == more_than(min, digit)
        assert digit.amt(2) == match_num(2, digit)
        assert digit.at_most(max) == at_most(max, digit)
        assert digit.between(min, max) == between(min, max, digit)
        assert digit.between(min, max, greedy=False) == between(min, max, digit, greedy=False)
        assert digit.between(min, max, possessive=True) == between(min, max, digit, possessive=True)
        assert digit.at_least_one() == at_least_one(digit)
        assert digit.at_least_one(greedy=False) == at_least_one(digit, greedy=False)
        assert digit.at_least_one(possessive=True) == at_least_one(digit, possessive=True)
        assert digit.at_least_none() == at_least_none(digit)
        assert digit.at_least_none(greedy=False) == at_least_none(digit, greedy=False)
        assert digit.at_least_none(possessive=True) == at_least_none(digit, possessive=True)
        # Order *does* matter
        assert digit.or_(input) == or_(digit, input)
        assert digit.either(input) == either(digit, input)
        assert digit.or_(input) == either(digit, input)

        assert options('multiline') + digit.group() == group(digit) + options('multiline')
        assert options('multiline') + digit.group(name='test') == group(digit, name='test') + options('multiline')
        assert options('multiline') + digit.if_not_preceded_by(input) == if_not_preceded_by(input) + options('multiline') + digit
        assert options('multiline') + digit.if_preceded_by(input) == if_preceded_by(input) + options('multiline') + digit
        assert options('multiline') + digit.if_not_proceded_by(input) == digit + if_not_proceded_by(input) + options('multiline')
        assert options('multiline') + digit.if_proceded_by(input) == digit + if_proceded_by(input) + options('multiline')
        assert options('multiline') + digit.if_enclosed_with('|') == if_enclosed_with('|', '|', digit) + options('multiline') == if_enclosed_with('|', pattern=digit) + options('multiline')
        assert options('multiline') + digit.at_least(min) == at_least(min, digit) + options('multiline')
        assert options('multiline') + digit.more_than(min) == more_than(min, digit) + options('multiline')
        assert options('multiline') + digit.amt(2) == match_num(2, digit) + options('multiline')
        assert options('multiline') + digit.at_most(max) == at_most(max, digit) + options('multiline')
        assert options('multiline') + digit.between(min, max) == between(min, max, digit) + options('multiline')
        assert options('multiline') + digit.between(min, max, greedy=False) == between(min, max, digit, greedy=False) + options('multiline')
        assert options('multiline') + digit.between(min, max, possessive=True) == between(min, max, digit, possessive=True) + options('multiline')
        assert options('multiline') + digit.at_least_one() == at_least_one(digit) + options('multiline')
        assert options('multiline') + digit.at_least_one(greedy=False) == at_least_one(digit, greedy=False) + options('multiline')
        assert options('multiline') + digit.at_least_one(possessive=True) == at_least_one(digit, possessive=True) + options('multiline')
        assert options('multiline') + digit.at_least_none() == at_least_none(digit) + options('multiline')
        assert options('multiline') + digit.at_least_none(greedy=False) == at_least_none(digit, greedy=False) + options('multiline')
        assert options('multiline') + digit.at_least_none(possessive=True) == at_least_none(digit, possessive=True) + options('multiline')
        assert options('multiline') + digit.or_(input) == or_(digit, input) + options('multiline')
        assert options('multiline') + digit.either(input) == either(digit, input) + options('multiline')
        assert options('multiline') + digit.or_(input) == either(digit, input) + options('multiline')
    except Exception as err:
        if 'require at least Python3.11' not in str(err):
            raise

    assert digit.optional == optional(digit)
    assert digit.repeat == repeat(digit)
    assert digit.exactly == is_exactly(digit)

    assert digit.append(input) == digit + input
    assert digit.prepend(input) == input + digit

    assert 'foo' + number + optional(whitespace) + word == number.append(whitespace.optional).prepend('foo').append(word)
    ans = r'(?:\s+)?((?:(?:a)+|b))(?=\w+)'
    assert (optional(whitespace) + group(either(repeat('a'), 'b')) + if_followed_by(word)).str() in ans
    assert whitespace.optional.append(literal('a').repeat.or_('b').group).if_followed_by(word).str() in ans
    assert (whitespace.optional + repeat('a').or_('b').group + if_followed_by(word)).str() in ans

def test_append_prepend():
    assert digit.opt.then(word).str() == r'(?:\d)?\w+'
    assert (digit + 'asdf').opt.then(word).str() == r'(?:\dasdf)?\w+'
    assert digit.opt.then(word).opt.then(word).str() == r'(?:(?:\d)?\w+)?\w+'
    assert digit.opt.word.opt.word.str() == r'(?:(?:\d)?\w+)?\w+'

    assert digit.opt.append(word).str() == r'(?:\d)?\w+'
    assert (digit + 'asdf').opt.append(word).str() == r'(?:\dasdf)?\w+'
    assert digit.opt.append(word).opt.append(word).str() == r'(?:(?:\d)?\w+)?\w+'
    assert digit.opt.word.opt.word.str() == r'(?:(?:\d)?\w+)?\w+'

    assert digit.prepend('asdf').str() == r'asdf\d'
    assert digit.prepend('asdf').prepend('1234').str() == r'1234asdf\d'
    assert digit.append('asdf').prepend('1234').str() == r'1234\dasdf'
    assert digit.append('asdf').prepend('1234').append('jkl;').str() == r'1234\dasdfjkl;'

    assert digit.append(whitespace.opt).str() == r'\d(?:\s+)?'
    assert (digit + whitespace.opt).str() == r'\d(?:\s+)?'
    assert digit.whitespace.opt.str() == r'(?:\d\s+)?'

    assert (digit + whitespace.optional).str() == str(digit + optional(whitespace))
    assert digit.whitespace.optional.str() == str(optional(digit.whitespace))

def test_no_duplicate_flags():
    r = lineStart + word + '/' + '/' + lineEnd
    assert r.str() == r'(?m)^\w+//$'

def test_accurate_types():
    # ...is a function
    assert type(ez.options) is type(test_accurate_types)
    assert type(ez.number) is ez.PythonEZRegex
    assert type(number) is ez.PythonEZRegex
    assert type(ez.python.number) is ez.PythonEZRegex
    assert type(ez.any_of) is ez.PythonEZRegex
    assert type(any_of) is ez.PythonEZRegex
    assert type(ez.python.any_of) is ez.PythonEZRegex
    assert type(ez.unicode) is ez.PythonEZRegex
    assert type(unicode) is ez.PythonEZRegex
    assert type(ez.python.unicode) is ez.PythonEZRegex
    assert type(ez.match_range) is ez.PythonEZRegex
    assert type(match_range) is ez.PythonEZRegex
    assert type(ez.python.match_range) is ez.PythonEZRegex
    assert type(ez.is_exactly) is ez.PythonEZRegex
    assert type(is_exactly) is ez.PythonEZRegex
    assert type(ez.python.is_exactly) is ez.PythonEZRegex
    assert type(ez.exactly) is ez.PythonEZRegex
    assert type(exactly) is ez.PythonEZRegex
    assert type(ez.python.exactly) is ez.PythonEZRegex
    assert type(ez.raw) is ez.PythonEZRegex
    assert type(raw) is ez.PythonEZRegex
    assert type(ez.python.raw) is ez.PythonEZRegex
    assert type(ez.match_amt) is ez.PythonEZRegex
    assert type(match_amt) is ez.PythonEZRegex
    assert type(ez.python.match_amt) is ez.PythonEZRegex
    assert type(ez.alpha) is ez.PythonEZRegex
    assert type(alpha) is ez.PythonEZRegex
    assert type(ez.python.alpha) is ez.PythonEZRegex
    assert type(ez.alphanum) is ez.PythonEZRegex
    assert type(alphanum) is ez.PythonEZRegex
    assert type(ez.python.alphanum) is ez.PythonEZRegex
    assert type(ez.letter) is ez.PythonEZRegex
    assert type(letter) is ez.PythonEZRegex
    assert type(ez.python.letter) is ez.PythonEZRegex
    assert type(ez.letter_num) is ez.PythonEZRegex
    assert type(letter_num) is ez.PythonEZRegex
    assert type(ez.python.letter_num) is ez.PythonEZRegex

def test_proper_sanitation():
    assert literal(r'\A').str() == r'\\A'
    assert raw(r'\A').str() == r'\A'
    assert raw(r'\A').word.str() == r'\A\w+'
    assert literal(r'\A').word.str() == r'\\A\w+'
    assert (digit + raw(r'\A').word).str() == r'\d\A\w+'
    assert (digit + literal(r'\A').word).str() == r'\d\\A\w+'
    assert digit.raw(r'\A').word.str() == r'\d\A\w+'
    assert digit.literal(r'\A').word.str() == r'\d\\A\w+'
    assert (digit.raw(r'\A') + word).str() == r'\d\A\w+'
    assert (digit.literal(r'\A') + word).str() == r'\d\\A\w+'

    # I'm not sure what this will do... it should raise an error?
    with pytest.raises(TypeError):
        digit.raw.word.str()
    # TODO: this could use some more tests in it

def test_replacements_not_interoperable():
    assert (rgroup(1) + 'foo').str() == r'\g<1>foo'
    assert (rgroup('bar') + 'foo').str() == r'\g<bar>foo'
    assert (rgroup('bar') + literal('foo')).str() == r'\g<bar>foo'
    assert (rgroup('bar') + raw('foo')).str() == r'\g<bar>foo'
    with pytest.raises(TypeError):
        (rgroup('bar') + word)


def test_all_parts_are_correct_type():
    for dialect in ALL_DIALECTS:
        for part_name in dialect.parts(include_functions=False):
            assert isinstance(getattr(dialect, part_name), EZRegex)
            assert isinstance(getattr(dialect, part_name), dialect)
        print(f'{dialect.__name__} passed')

# Wouldn't it be awkward if these failed
def test_README_examples(capsys):
    assert ('foo' + number + optional(whitespace) + group(word)).str() == r'foo\d+(?:\s+)?(\w+)'
    # Or if you prefer the method syntax (they can be mixed)
    assert number.append(whitespace.optional).prepend('foo').append(word.group()).str() == r'foo\d+(?:\s+)?(\w+)'

    # These match `foo123abc` and `foo123 abc`
    # but not `abc123foo` or  `foo bar`



    import ezregex as ez

    # ow is part of ez already as "optional chunk of whitespace" (`\s*`)
    params = ez.group(ez.at_least_none(ez.ow + ez.word + ez.ow + ez.optional(',') + ez.ow))
    # Seperate parts as variables for cleaner patterns
    function = ez.word + ez.ow + '(' + params + ')'

    assert function.search('some string containing func( param1 , param2)')

    # Boolean test
    assert 'some string containing func( param1 , param2)' in function

    test_ouptut = r"""\
╭─────────────────────────────── Testing Regex ────────────────────────────────╮
│ Testing expression:                                                          │
│         \w+\s*\(((?:\s*\w+\s*,?\s*)*)\)                                      │
│ for matches in:                                                              │
│         this should match func(param1,  param2 ), foo(), and bar( foo,)      │
│                                                                              │
│ Match = "func(param1,   param2 )" (18:39)                                    │
│ Unnamed Groups:                                                              │
│         1: "param1,     param2 " (23:38)                                     │
│                                                                              │
│ Match = "foo()" (41:46)                                                      │
│ Unnamed Groups:                                                              │
│         1: "" (45:45)                                                        │
│                                                                              │
│ Match = "bar( foo,)" (52:62)                                                 │
│ Unnamed Groups:                                                              │
│         1: " foo," (56:61)                                                   │
│                                                                              │
│                                                                              │
╰─────────────────────────────────── Found  ───────────────────────────────────╯
"""

    # The test() method is helpful for debugging, and color codes groups for you
    assert function.test('this should match func(param1,\tparam2 ), foo(), and bar( foo,)')
    captured = capsys.readouterr().out
    # it's hard to test whole, because the width is dynamic to the terminal size
    assert 'Testing expression:' in captured
    assert 'for matches in:' in captured
    assert 'Match = "func(param1,   param2 )" (18:39)' in captured
    assert 'Unnamed Groups:' in captured
    assert 'Match = "foo()" (41:46)' in captured
    assert 'Unnamed Groups:' in captured
    assert 'Match = "bar( foo,)" (52:62)' in captured
    assert 'Unnamed Groups:' in captured



    # Element functions
    assert (
        optional(whitespace) + group(either(repeat('a'), 'b')) + if_followed_by(word) ==
        # Elemental methods
        whitespace.optional.append(literal('a').repeat.or_('b').group).if_followed_by(word) ==
        # Mixed
        whitespace.optional + repeat('a').or_('b').group + if_followed_by(word)
    )





    import ezregex as ez # The python dialect is the defualt dialect
    assert repr(ez.group(digit, name='name') + ez.earlier_group('name')) == r"PythonEZRegex((?P<name>\d)(?P=name), {'_compiled': None, 'flags': set(), 'replacement': False})"
    import ezregex.javascript as ez
    assert repr(ez.group(digit, name='name') + ez.earlier_group('name')) == r"JavascriptEZRegex(/(?<name>\d)\k<name>/, {'_string_anchor_used': False, 'flags': set(), 'replacement': False})"


common_flag_docs = {
    'ascii': '''Make matching words, word boundaries, digits, and whitespace \
perform ASCII-only matching instead of full Unicode matching (which is \
default). This is only meaningful for Unicode (str) patterns, and is \
ignored for bytes patterns''',
    'ignore_case': '''Perform case-insensitive matching, including expressions that \
explicitly use uppercase members. Full Unicode matching (such as Ü \
matching ü) also works unless the ASCII flag is used to disable \
non-ASCII matches. The current locale does not change the effect of \
this flag unless the LOCALE flag is also used''',
    'unicode': '''Match using the full unicode standard, instead of just ASCII. \
Enabled by default, and therefore redundant''',
    'global': '''Global mode. Match everything in the given string, instead of just the first match''',
    'anchor': '''The pattern is forced to become anchored at the start of the \
search or at the position of the last successful match.''',
    'single_line': '''Not recommended. Makes the '.' special character match any character at all, including \
a newline. It's recommended you simply use literally_anything instead''',
    'multiline': '''Not recommended. Makes the '^' and '$' special characters match the start and end of \
lines, instead of the start and end of the string. This is automatically inserted when using \
line_start and line_end, you shouldn\'t need to add it manually''',
    'verbose': '''Not recommended. Allows for comments and whitespace, which both don\'t do anything \
in this library.''',
    'lazy': '''The engine will per default to lazy matching, instead of greedy. It's recommended you \
just specify greedy=False instead''',
    'duplicate_groups': '''This allows regex to accept duplicate pattern names, \
however each capture group still has its own ID. Thus the two capture \
groups produce their own match instead of a single combined one''',
}


#!/usr/bin/env python3
from pathlib import Path
import shutil
from ezregex.psuedonyms import psuedonyms
from ezregex.EZRegex import _to_camel_case
import ezregex as ez

# from pdoc import pdoc
# from pdoc import render

here = Path(__file__).parent
dialects = here / "docs" / "dialects"

if dialects.exists():
    shutil.rmtree(dialects)

dialects.mkdir(parents=True, exist_ok=True)

# # Render parts of pdoc's documentation into docs/api...
# render.configure(template_directory=here / "docs" / "pdoc-template", search=False)

# I gave up trying to mess with pdoc and just decided to parse everything by hand for now

# TODO: this still has issues:
# I think some of the replacement regexs aren't getting added?
# General formatting (still want to see if I can get pdoc to work with it)
# Overwriting docs should be a thing (not a problem with this specifically, but elsewhere)
# There's no args, or arg formatting yet. That both needs to be added here, and in all the docstrings
for dialect in [ez.r.REZRegex, ez.javascript.JavascriptEZRegex, ez.pcre2.PCRE2EZRegex, ez.python.PythonEZRegex]:
    doc = ''
    doc += f"# {dialect.__name__}\n\n{dialect.__doc__}\n\n"

    doc += f"## options\n\n{dialect.options.__doc__}\n\n"

    for repl in (False, True):
        for part_name in dialect.parts(include_psuedonyms=False, include_functions=False):
            if getattr(dialect, part_name).replacement != repl:
                continue
            doc += f"## {part_name}\n"
            aliases = set()
            for i in psuedonyms.get(part_name, set()):
                aliases.add(_to_camel_case(i))
                aliases.add(i)
            aliases.add(part_name)
            aliases.add(_to_camel_case(part_name))
            aliases.remove(part_name)
            doc += f"<span style='font-style: italic; font-size: small;'>Aliases: {', '.join(aliases)}</span>\n\n"
            if (d := getattr(dialect, part_name).docstring):
                doc += f"{d}\n\n"
        if not repl:
            doc += '\n\n---\n\n## Replacement EZRegexs\n\n'
    # Add docs for the replace() function at the end
    doc += f"## replace\n\n{dialect.replace.__doc__}\n\n"

    (dialects / f"{dialect.__name__}.md").write_text(doc)



# import subprocess
# subprocess.run(["gendocs", "--config", "mkgendocs.yml"])


#!/usr/bin/env python3
""" Support for the Python dialect of regular expressions"""
__version__ = '2.1.0'

import re
import sys

from .. import EZRegex
from ..mixins import (AdvancedGroupsMixin, BaseMixin, AssertionsMixin, GroupsMixin, AnchorsMixin, ReplacementsMixin)
from ..flag_docs import common_flag_docs
from sys import version_info

# TODO: make all the flag functions here also accept re.FLAG types (internally they should work the same though)

class PythonEZRegex(
    BaseMixin(allow_greedy=True, allow_possessive=version_info >= (3, 11)),
    AssertionsMixin(),
    GroupsMixin(),
    AdvancedGroupsMixin(),
    # only 3.14+ has support for \z
    AnchorsMixin(word=False, string_end=r'\Z'),
    ReplacementsMixin(
        named_group=lambda name, cur=...: fr'{cur}\g<{name}>',
        numbered_group=lambda num, cur=...: fr'{cur}\g<{num}>'
    ),
    EZRegex,

    escape_chars=b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f',
    flags={
        'ascii': 'a',
        'ignore_case': 'i',
        'single_line': 's',
        'locale': 'L',
        'multiline': 'm',
        'unicode': 'u'
    },
    flags_docs_map={**common_flag_docs, 'locale': '''Try not to use this, and rely on unicode matching instead'''},
    flags_docs_link='https://docs.python.org/3/library/re.html#flags',

    # If we try to combine it with something, invalidate the cache
    variables={
        '_compiled': (None, lambda l, r: None),
    }
):
    """
    Official docs:
    https://docs.python.org/3/library/re.html
    """

    # TODO: rewrite these using existing elements, and put them in a premade mixin, so they can be generic
    # Source: http://stackoverflow.com/questions/201323/ddg#201378
    email = r"(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|\"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*\")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9]))\.){3}(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])"
    "Matches an email"
    # Source: https://semver.org/ (at the bottom)
    version = r"(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"
    """The *official* regex for matching version numbers from https://semver.org/. It includes 5 groups that can be
    matched/replaced: `major`, `minor`, `patch`, `prerelease`, and `buildmetadata`"""
    version_numbered = r"(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"
    "Same as `version`, but it uses numbered groups for each version number instead of named groups"

    @EZRegex.exclude
    def compile(self, add_flags=True):
        return re.compile(self._compile(add_flags=add_flags))

    @property
    @EZRegex.exclude
    def compiled(self):
        if self._compiled is None:
            self.__dict__['_compiled'] = self.compile()
        return self._compiled

    # Shadowing the re functions
    @EZRegex.exclude
    def search(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().search(string, pos, endpos)

    @EZRegex.exclude
    def match(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().match(string, pos, endpos)

    @EZRegex.exclude
    def fullmatch(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().fullmatch(string, pos, endpos)

    @EZRegex.exclude
    def split(self, string, maxsplit=0):
        return self.compile().split(string, maxsplit)

    @EZRegex.exclude
    def findall(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().findall(string, pos, endpos)

    @EZRegex.exclude
    def finditer(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().finditer(string, pos, endpos)

    @EZRegex.exclude
    def sub(self, repl, string, count=0):
        return self.compile().sub(repl, string, count)

    @EZRegex.exclude
    def subn(self, repl, string, count=0):
        return self.compile().subn(repl, string, count)



""" Support for the R dialect of regular expressions"""
__version__ = '1.1.0'

import re
from .. import EZRegex
from ..mixins import (BaseMixin, GroupsMixin, AnchorsMixin, ReplacementsMixin, _parse_any_of_params)

class REZRegex(
    BaseMixin(allow_greedy=False, allow_possessive=False),
    GroupsMixin(named_group=None),
    AnchorsMixin(string=False),
    ReplacementsMixin(
        numbered_group=lambda num, cur=...: fr'{cur}\{num}',
        named_group=None,
    ),
    EZRegex,

    escape_chars=b'|()[{^$*+?.-',
    flags={},
    flags_docs_map={},
):
    """
    Official docs:
    https://www.rdocumentation.org/packages/base/versions/3.6.2/topics/regex
    """

    def _final_func(self, s:str) -> str:
        # Double escape all backslashes
        s = s.replace('\\', '\\\\')
        # Things which are control characters should *not* be double escaped
        s = re.sub(r'\\\\([abtnvfrxuU])', r'\\\g<1>', s)
        return s

    punctuation = r'[]`~!@#\$%^&\*\(\)\-_=\+[\{}\\\|;:\'",<\.>/\?]'

    def any_of(*inputs, chars=None, split=None, cur=...):
        chars, inputs = _parse_any_of_params(*inputs, chars=chars, split=split)

        if chars:
            if ']' in inputs:
                # You can't escape ] in an R character class, it has to be first
                inputs.remove(']')
                inputs.insert(0, ']')
            cur += r'['
            for i in inputs:
                cur += i
            cur += r']'
        else:
            cur += r'(?:'
            for i in inputs:
                cur += i
                cur += '|'
            cur = cur[:-1]
            cur += r')'
        return cur

    # For some reason \N doesn't seem to work? It doesn't matter that much
    chunk = r'[^\n]+'

    # Matches any single character except line break characters, like the dot, but is not affected by any options that make the dot match all characters including line breaks.
    not_newline = r'\N'

    # def options(*args, **kwargs):
    #     raise ValueError('Flags are not supported in R dialect')
